"""
AMRVAC-specific IO functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
import numpy as np
from yt.utilities.io_handler import \
    BaseIOHandler
from yt.geometry.selection_routines import \
    GridSelector
from .datfile_utils import get_single_block_field_data


class AMRVACIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = 'amrvac'

    def __init__(self, ds):
        BaseIOHandler.__init__(self, ds)
        self.ds = ds
        self.datfile = ds.parameter_filename
        header = self.ds.parameters
        self.block_shape = np.append(header["block_nx"], header["nw"])

    def _read_particle_coords(self, chunks, ptf):
        """Not implemented yet."""
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        raise NotImplementedError

    def _read_particle_fields(self, chunks, ptf, selector):
        """Not implemented yet."""
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        raise NotImplementedError

    def _read_data(self, grid, field):
        """Retrieve field data from a grid.

        Parameters
        ----------
        grid : yt.frontends.amrvac.data_structures.AMRVACGrid
            The grid from which data is to be read.
        field : str
            A field name.

        Returns
        -------
        data : np.ndarray
            A 3D array of float64 type representing grid data.

        """
        ileaf = grid.id
        offset = grid._index.block_offsets[ileaf]
        field_idx = self.ds.parameters['w_names'].index(field)
        with open(self.datfile, "rb") as istream:
            data = get_single_block_field_data(istream, offset, self.block_shape, field_idx)

        # Always convert data to 3D, as grid.ActiveDimensions is always 3D
        while len(data.shape) < 3:
            data = data[..., np.newaxis]
        return data

    def _read_fluid_selection(self, chunks, selector, fields, size):
        """Retrieve field(s) data in a selected region of space.

        Parameters
        ----------
        chunks : generator
            A generator for multiple chunks, each of which contains a list of grids.

        selector : yt.geometry.selection_routines.SelectorObject
            A spatial region selector.

        fields : list
            A list of tuples (ftype, fname).

        size : np.int64
            The cumulative number of objs contained in all chunks.

        Returns
        -------
        data_dict : dict
            keys are the (ftype, fname) tuples and values are arrays that have been masked using
            whatever selector method is appropriate. Arrays have dtype float64. 
        """

        #@Notes from Niels:
        #The chunks list has YTDataChunk objects containing the different grids.
        #The list of grids can be obtained by doing eg. grids_list = chunks[0].objs or chunks[1].objs etc.
        #Every element in "grids_list" is then an AMRVACGrid object, and has hence all attributes of a grid
        #    (Level, ActiveDimensions, LeftEdge, etc.)

        chunks = list(chunks)
        data_dict = {} # <- return variable
        if isinstance(selector, GridSelector):
            if not len(chunks) == len(chunks[0].objs) == 1:
                raise RuntimeError
            grid = chunks[0].objs[0]
            for ftype, fname in fields:
                data_dict[ftype, fname] = self._read_data(grid, fname)
        else:
            if size is None:
                size = sum((g.count(selector) for chunk in chunks
                            for g in chunk.objs))

            for field in fields:
                data_dict[field] = np.empty(size, dtype='float64')

            #nb_grids = sum(len(chunk.objs) for chunk in chunks)

            ind = 0
            for chunk in chunks:
                for grid in chunk.objs:
                    nd = 0
                    for field in fields:
                        ftype, fname = field
                        data = self._read_data(grid, fname)
                        nd = grid.select(selector, data, data_dict[field], ind)
                    ind += nd

        return data_dict


    def _read_chunk_data(self, chunk, fields):
        """Not implemented yet."""
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        # it should be used by _read_fluid_selection instead of _read_data
        raise NotImplementedError



"""
Derived field base class.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import contextlib
import inspect
import re
import warnings

from yt.extern.six import string_types, PY2
from yt.funcs import \
    ensure_list
from .field_exceptions import \
    NeedsGridType, \
    NeedsOriginalGrid, \
    NeedsDataField, \
    NeedsProperty, \
    NeedsParameter, \
    FieldUnitsError
from .field_detector import \
    FieldDetector
from yt.units.unit_object import \
    Unit
import yt.units.dimensions as ytdims
from yt.utilities.exceptions import \
    YTFieldNotFound


def TranslationFunc(field_name):
    def _TranslationFunc(field, data):
        # We do a bunch of in-place modifications, so we will copy this.
        return data[field_name].copy()
    _TranslationFunc.alias_name = field_name
    return _TranslationFunc

def NullFunc(field, data):
    raise YTFieldNotFound(field.name)

class DerivedField(object):
    """
    This is the base class used to describe a cell-by-cell derived field.

    Parameters
    ----------

    name : str
       is the name of the field.
    function : callable
       A function handle that defines the field.  Should accept
       arguments (field, data)
    units : str
       A plain text string encoding the unit, or a query to a unit system of
       a dataset. Powers must be in python syntax (** instead of ^). If set
       to "auto" the units will be inferred from the units of the return
       value of the field function, and the dimensions keyword must also be
       set (see below).
    take_log : bool
       Describes whether the field should be logged
    validators : list
       A list of :class:`FieldValidator` objects
    sampling_type : string, default = "cell"
        How is the field sampled?  This can be one of the following options at
        present: "cell" (cell-centered), "discrete" (or "particle") for
        discretely sampled data.
    particle_type : bool
       (Deprecated) Is this a particle (1D) field?  This is deprecated. Use
       sampling_type = "discrete" or sampling_type = "particle".  This will
       *override* sampling_type.
    vector_field : bool
       Describes the dimensionality of the field.  Currently unused.
    display_field : bool
       Governs its appearance in the dropdowns in Reason
    not_in_all : bool
       Used for baryon fields from the data that are not in all the grids
    display_name : str
       A name used in the plots
    output_units : str
       For fields that exist on disk, which we may want to convert to other
       fields or that get aliased to themselves, we can specify a different
       desired output unit than the unit found on disk.
    dimensions : str or object from yt.units.dimensions
       The dimensions of the field, only needed if units="auto" and only used
       for error checking.
    nodal_flag : array-like with three components
       This describes how the field is centered within a cell. If nodal_flag
       is [0, 0, 0], then the field is cell-centered. If any of the components
       of nodal_flag are 1, then the field is nodal in that direction, meaning
       it is defined at the lo and hi sides of the cell rather than at the center.
       For example, a field with nodal_flag = [1, 0, 0] would be defined at the
       middle of the 2 x-faces of each cell. nodal_flag = [0, 1, 1] would mean the
       that the field defined at the centers of the 4 edges that are normal to the
       x axis, while nodal_flag = [1, 1, 1] would be defined at the 8 cell corners.
    """
    _inherited_particle_filter = False
    def __init__(self, name, sampling_type, function, units=None,
                 take_log=True, validators=None,
                 particle_type=None, vector_field=False, display_field=True,
                 not_in_all=False, display_name=None, output_units=None,
                 dimensions=None, ds=None, nodal_flag=None):
        self.name = name
        self.take_log = take_log
        self.display_name = display_name
        self.not_in_all = not_in_all
        self.display_field = display_field
        if particle_type is True:
            warnings.warn("particle_type for derived fields "
                          "has been replaced with sampling_type = 'particle'",
                          DeprecationWarning)
            sampling_type = "particle"
        self.sampling_type = sampling_type
        self.vector_field = vector_field
        self.ds = ds

        if self.ds is not None:
            self._ionization_label_format = self.ds._ionization_label_format
        else:
            self._ionization_label_format = "roman_numeral"

        if nodal_flag is None:
            self.nodal_flag = [0, 0, 0]
        else:
            self.nodal_flag = nodal_flag

        self._function = function

        if validators:
            self.validators = ensure_list(validators)
        else:
            self.validators = []

        # handle units
        if units is None:
            self.units = ''
        elif isinstance(units, string_types):
            if units.lower() == 'auto':
                if dimensions is None:
                    raise RuntimeError("To set units='auto', please specify the dimensions "
                                       "of the field with dimensions=<dimensions of field>!")
                self.units = None
            else:
                self.units = units
        elif isinstance(units, Unit):
            self.units = str(units)
        elif isinstance(units, bytes):
            self.units = units.decode("utf-8")
        else:
            raise FieldUnitsError("Cannot handle units '%s' (type %s)." \
                                  "Please provide a string or Unit " \
                                  "object." % (units, type(units)) )
        if output_units is None:
            output_units = self.units
        self.output_units = output_units

        if isinstance(dimensions, string_types):
            dimensions = getattr(ytdims, dimensions)
        self.dimensions = dimensions

    def _copy_def(self):
        dd = {}
        dd['name'] = self.name
        dd['units'] = self.units
        dd['take_log'] = self.take_log
        dd['validators'] = list(self.validators)
        dd['sampling_type'] = self.sampling_type
        dd['vector_field'] = self.vector_field
        dd['display_field'] = True
        dd['not_in_all'] = self.not_in_all
        dd['display_name'] = self.display_name
        return dd

    @property
    def particle_type(self):
        return self.sampling_type in ("discrete", "particle")

    def get_units(self):
        if self.ds is not None:
            u = Unit(self.units, registry=self.ds.unit_registry)
        else:
            u = Unit(self.units)
        return u.latex_representation()

    def get_projected_units(self):
        if self.ds is not None:
            u = Unit(self.units, registry=self.ds.unit_registry)
        else:
            u = Unit(self.units)
        return (u*Unit('cm')).latex_representation()

    def check_available(self, data):
        """
        This raises an exception of the appropriate type if the set of
        validation mechanisms are not met, and otherwise returns True.
        """
        for validator in self.validators:
            validator(data)
        # If we don't get an exception, we're good to go
        return True

    def get_dependencies(self, *args, **kwargs):
        """
        This returns a list of names of fields that this field depends on.
        """
        e = FieldDetector(*args, **kwargs)
        if self._function.__name__ == '<lambda>':
            e.requested.append(self.name)
        else:
            e[self.name]
        return e

    def _get_needed_parameters(self, fd):
        params = []
        values = []
        permute_params = {}
        vals = [v for v in self.validators if isinstance(v, ValidateParameter)]
        for val in vals:
            if val.parameter_values is not None:
                permute_params.update(val.parameter_values)
            else:
                params.extend(val.parameters)
                values.extend(
                    [fd.get_field_parameter(fp) for fp in val.parameters])
        return dict(zip(params, values)), permute_params

    _unit_registry = None
    @contextlib.contextmanager
    def unit_registry(self, data):
        old_registry = self._unit_registry
        if hasattr(data, 'unit_registry'):
            ur = data.unit_registry
        elif hasattr(data, 'ds'):
            ur = data.ds.unit_registry
        else:
            ur = None
        self._unit_registry = ur
        yield
        self._unit_registry = old_registry

    def __call__(self, data):
        """ Return the value of the field in a given *data* object. """
        self.check_available(data)
        original_fields = data.keys() # Copy
        if self._function is NullFunc:
            raise RuntimeError(
                "Something has gone terribly wrong, _function is NullFunc " +
                "for %s" % (self.name,))
        with self.unit_registry(data):
            dd = self._function(self, data)
        for field_name in data.keys():
            if field_name not in original_fields:
                del data[field_name]
        return dd

    def get_source(self):
        """
        Return a string containing the source of the function (if possible.)
        """
        return inspect.getsource(self._function)

    def get_label(self, projected=False):
        """
        Return a data label for the given field, including units.
        """
        name = self.name[1]
        if self.display_name is not None:
            name = self.display_name

        # Start with the field name
        data_label = r"$\rm{%s}" % name

        # Grab the correct units
        if projected:
            raise NotImplementedError
        else:
            if self.ds is not None:
                units = Unit(self.units, registry=self.ds.unit_registry)
            else:
                units = Unit(self.units)
        # Add unit label
        if not units.is_dimensionless:
            data_label += r"\ \ (%s)" % (units.latex_representation())

        data_label += r"$"
        return data_label

    def __repr__(self):
        if PY2:
            func_name = self._function.func_name
        else:
            func_name = self._function.__name__
        if self._function == NullFunc:
            s = "On-Disk Field "
        elif func_name == "_TranslationFunc":
            s = "Alias Field for \"%s\" " % (self._function.alias_name,)
        else:
            s = "Derived Field "
        if isinstance(self.name, tuple):
            s += "(%s, %s): " % self.name
        else:
            s += "%s: " % (self.name)
        s += "(units: %s" % self.units
        if self.display_name is not None:
            s += ", display_name: '%s'" % (self.display_name)
        if self.particle_type:
            s += ", particle field"
        s += ")"
        return s

    def _is_ion(self):
        p = re.compile("_p[0-9]+_")
        result = False
        if p.search(self.name[1]) is not None:
            result = True
        return result

    def _ion_to_label(self):
        #check to see if the output format has changed
        if self.ds is not None:
            self._ionization_label_format = self.ds._ionization_label_format

        pnum2rom = {
            "0":"I", "1":"II", "2":"III", "3":"IV", "4":"V",
            "5":"VI", "6":"VII", "7":"VIII", "8":"IX", "9":"X",
            "10":"XI", "11":"XII", "12":"XIII", "13":"XIV", "14":"XV",
            "15":"XVI", "16":"XVII", "17":"XVIII", "18":"XIX", "19":"XX"}

        # first look for charge to decide if it is an ion
        p = re.compile("_p[0-9]+_")
        m = p.search(self.name[1])
        if m is not None:

            # Find the ionization state
            pstr = m.string[m.start()+1:m.end()-1]
            segments = self.name[1].split("_")

            # find the ionization index
            for i,s in enumerate(segments):
                if s == pstr: ipstr = i

            
            for i,s in enumerate(segments):
                #If its the species we don't want to change the capitalization
                if i == ipstr -1: continue
                segments[i] = s.capitalize()

            species = segments[ipstr-1]

            #If there is a number in the species part of the label
            #that indicates part of a molecule
            symbols = []
            for symb in species:
                # can't just use underscore b/c gets replaced later with space
                if symb.isdigit(): symbols.append("latexsub{"+symb+'}')
                else: symbols.append(symb)
            species_label =  "".join(symbols)


            # Use roman numerals for ionization

            if self._ionization_label_format == "roman_numeral":
                roman = pnum2rom[pstr[1:]]
                label = species_label + '\ ' + roman + '\ ' + \
                    '\ '.join(segments[ipstr+1:])

            # use +/- for ionization
            else: 
                sign = '+'*int(pstr[1:])
                label ='{' +species_label+ '}'+ '^{'+sign+'}' + '\ ' + \
                    '\ '.join(segments[ipstr+1:])

        else:
            label = self.name[1]
        return label

    def get_latex_display_name(self):
        label = self.display_name
        if label is None:
            if self._is_ion():
                fname = self._ion_to_label()
                label = r'$\rm{'+fname.replace('_','\ ')+r'}$'
                label = label.replace('latexsub', '_')
            else:
                label = r'$\rm{'+self.name[1].replace('_','\ ').title()+r'}$'
        elif label.find('$') == -1:
            label = label.replace(' ','\ ')
            label = r'$\rm{'+label+r'}$'
        return label


class FieldValidator(object):
    pass

class ValidateParameter(FieldValidator):
    def __init__(self, parameters, parameter_values=None):
        """
        This validator ensures that the dataset has a given parameter.

        If *parameter_values* is supplied, this will also ensure that the field
        is available for all permutations of the field parameter.
        """
        FieldValidator.__init__(self)
        self.parameters = ensure_list(parameters)
        self.parameter_values = parameter_values
    def __call__(self, data):
        doesnt_have = []
        for p in self.parameters:
            if not data.has_field_parameter(p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsParameter(doesnt_have)
        return True

class ValidateDataField(FieldValidator):
    def __init__(self, field):
        """
        This validator ensures that the output file has a given data field stored
        in it.
        """
        FieldValidator.__init__(self)
        self.fields = ensure_list(field)
    def __call__(self, data):
        doesnt_have = []
        if isinstance(data, FieldDetector): return True
        for f in self.fields:
            if f not in data.index.field_list:
                doesnt_have.append(f)
        if len(doesnt_have) > 0:
            raise NeedsDataField(doesnt_have)
        return True

class ValidateProperty(FieldValidator):
    def __init__(self, prop):
        """
        This validator ensures that the data object has a given python attribute.
        """
        FieldValidator.__init__(self)
        self.prop = ensure_list(prop)
    def __call__(self, data):
        doesnt_have = []
        for p in self.prop:
            if not hasattr(data,p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsProperty(doesnt_have)
        return True

class ValidateSpatial(FieldValidator):
    def __init__(self, ghost_zones = 0, fields=None):
        """
        This validator ensures that the data handed to the field is of spatial
        nature -- that is to say, 3-D.
        """
        FieldValidator.__init__(self)
        self.ghost_zones = ghost_zones
        self.fields = fields

    def __call__(self, data):
        # When we say spatial information, we really mean
        # that it has a three-dimensional data structure
        #if isinstance(data, FieldDetector): return True
        if not getattr(data, '_spatial', False):
            raise NeedsGridType(self.ghost_zones,self.fields)
        if self.ghost_zones <= data._num_ghost_zones:
            return True
        raise NeedsGridType(self.ghost_zones,self.fields)

class ValidateGridType(FieldValidator):
    def __init__(self):
        """
        This validator ensures that the data handed to the field is an actual
        grid patch, not a covering grid of any kind.
        """
        FieldValidator.__init__(self)
    def __call__(self, data):
        # We need to make sure that it's an actual AMR grid
        if isinstance(data, FieldDetector): return True
        if getattr(data, "_type_name", None) == 'grid': return True
        raise NeedsOriginalGrid()



"""
Enzo-P misc tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.frontends.enzo_p.misc import \
    get_block_info, \
    get_root_blocks, \
    get_root_block_id, \
    is_parent, \
    nested_dict_get

def get_random_block_string(max_n=64, random_state=None, level=None):
    if max_n == 1:
        assert level is None or level == 0
        return 0, 0, 'B'
    elif max_n < 1:
        raise ValueError('max_n must be a positive integer')

    if random_state is None:
        random_state = np.random.RandomState()

    max_l = int(np.log2(max_n))
    form = "%0" + str(max_l) + "d"
    num10 = random_state.randint(0, high=max_n)
    num2 = form % int(bin(num10)[2:]) #the slice clips the '0b' prefix

    if level is None:
        level = random_state.randint(0, high=max_l)
    if level > 0:
        my_block = "%s:%s" % (num2[:-level], num2[-level:])
    else:
        my_block = num2
    my_block = "B" + my_block

    return num10, level, my_block

def flip_random_block_bit(block, rs):
    """
    Flips a bit string in one of the block descriptors in a given block name
    """
    # split block into descriptors for each dimension
    descriptors = block[1:].split('_')

    # choose which descriptor to modify
    flippable = [i for i,descr in enumerate(descriptors) if len(descr) > 0]
    if len(flippable) == 0: # when block in ['B', 'B_', 'B__']
        raise ValueError("%s has no bits that can be flipped" % block)
    descr_index = flippable[rs.randint(0,len(flippable))]

    # split block descriptor into left and right parts
    parts = descriptors[descr_index].split(':')
    # select the part to be modified
    if len(parts) == 1: # block is unrefined
        part_index = 0
    elif len(parts[0]) == 0: # The root block index can't be modified
        part_index = 1
    else:
        part_index = rs.randint(0, high = 2)
    modify_part = parts[part_index]

    # flip a bit in modify_part, and return the new block name with this change
    flip_index = rs.randint(0, high = len(modify_part))
    parts[part_index] = "%s%d%s" % (modify_part[:flip_index],
                                    (int(modify_part[flip_index]) + 1) % 2,
                                    modify_part[flip_index+1:])
    descriptors[descr_index] = ':'.join(parts)
    return 'B' + '_'.join(descriptors)

def test_get_block_info():
    rs = np.random.RandomState(45047)
    max_n = 64
    for i in range(10):
        n, l, b = get_random_block_string(
            max_n=max_n, random_state=rs)
        level, left, right = get_block_info(b, min_dim=1)
        assert level == l
        assert left == float(n) / max_n
        assert right == float(n+1) / max_n

    for block in ['B', 'B_', 'B__']:
        level, left, right = get_block_info(block)
        assert level == 0
        assert (left == 0.0).all()
        assert (right == 1.0).all()

def test_root_blocks():
    rs = np.random.RandomState(45652)
    for i in range(6):
        max_n = 2**i
        n1, l1, b1 = get_random_block_string(
            max_n=max_n, random_state=rs, level=0)
        n2, l2, b2 = get_random_block_string(
            max_n=32, random_state=rs, level=0)
        block = "%s:%s" % (b1, b2[1:])

        nrb = get_root_blocks(block, min_dim=1)
        assert nrb == max_n
        rbid = get_root_block_id(block, min_dim=1)
        assert rbid == n1

def test_is_parent():
    rs = np.random.RandomState(45652)
    for dim in [1,2,3]:
        for i in range(6):
            max_n = 2**i

            descriptors = []
            for j in range(dim):
                n1, l1, b1 = get_random_block_string(
                    max_n=max_n, random_state=rs, level=0)
                n2, l2, b2 = get_random_block_string(
                    max_n=32, random_state=rs, level=0)
                descriptors.append("%s:%s" % (b1[1:], b2[1:]))
            block = "B" + "_".join(descriptors)
            # since b2 is computed with max_n=32 in the for-loop, block always
            # has a refined great-great-grandparent
            parent      = "B" + "_".join([elem[:-1] for elem in descriptors])
            grandparent = "B" + "_".join([elem[:-2] for elem in descriptors])

            assert is_parent(parent,block)
            assert is_parent(grandparent, parent)
            assert not is_parent(grandparent, block)
            assert not is_parent(block, parent)
            assert not is_parent(flip_random_block_bit(parent, rs), block)

def test_nested_dict_get():
    rs = np.random.RandomState(47988)
    keys = []
    my_dict = None
    for i in range(5):
        k = str(rs.randint(0, high=1000000))
        if my_dict is None:
            v = str(rs.randint(0, high=1000000))
            keys.append(v)
            my_dict = {k: v}
        else:
            my_dict = {k: my_dict}
        keys.append(k)
    keys.reverse()
    assert nested_dict_get(my_dict, keys[:-1]) == keys[-1]

    my_def = "devron"
    assert nested_dict_get(
        my_dict, "system", default=my_def) == my_def



"""
Geometry container base class.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
from yt.extern.six.moves import cPickle
import weakref
from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np

from yt.config import ytcfg
from yt.funcs import iterable
from yt.units.yt_array import \
    YTArray, uconcatenate
from yt.utilities.io_handler import io_registry
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface, parallel_root_only
from yt.utilities.exceptions import YTFieldNotFound

class Index(ParallelAnalysisInterface):
    """The base index class"""
    _global_mesh = True
    _unsupported_objects = ()
    _index_properties = ()

    def __init__(self, ds, dataset_type):
        ParallelAnalysisInterface.__init__(self)
        self.dataset = weakref.proxy(ds)
        self.ds = self.dataset

        self._initialize_state_variables()

        mylog.debug("Initializing data storage.")
        self._initialize_data_storage()

        mylog.debug("Setting up domain geometry.")
        self._setup_geometry()

        mylog.debug("Initializing data grid data IO")
        self._setup_data_io()

        # Note that this falls under the "geometry" object since it's
        # potentially quite expensive, and should be done with the indexing.
        mylog.debug("Detecting fields.")
        self._detect_output_fields()

    def _initialize_state_variables(self):
        self._parallel_locking = False
        self._data_file = None
        self._data_mode = None
        self.num_grids = None

    def _initialize_data_storage(self):
        if not ytcfg.getboolean('yt','serialize'): return
        fn = self.ds.storage_filename
        if fn is None:
            if os.path.isfile(os.path.join(self.directory,
                                "%s.yt" % self.ds.unique_identifier)):
                fn = os.path.join(self.directory,"%s.yt" % self.ds.unique_identifier)
            else:
                fn = os.path.join(self.directory,
                        "%s.yt" % self.dataset.basename)
        dir_to_check = os.path.dirname(fn)
        if dir_to_check == '':
            dir_to_check = '.'
        # We have four options:
        #    Writeable, does not exist      : create, open as append
        #    Writeable, does exist          : open as append
        #    Not writeable, does not exist  : do not attempt to open
        #    Not writeable, does exist      : open as read-only
        exists = os.path.isfile(fn)
        if not exists:
            writeable = os.access(dir_to_check, os.W_OK)
        else:
            writeable = os.access(fn, os.W_OK)
        writeable = writeable and not ytcfg.getboolean('yt','onlydeserialize')
        # We now have our conditional stuff
        self.comm.barrier()
        if not writeable and not exists: return
        if writeable:
            try:
                if not exists: self.__create_data_file(fn)
                self._data_mode = 'a'
            except IOError:
                self._data_mode = None
                return
        else:
            self._data_mode = 'r'

        self.__data_filename = fn
        self._data_file = h5py.File(fn, self._data_mode)

    def __create_data_file(self, fn):
        # Note that this used to be parallel_root_only; it no longer is,
        # because we have better logic to decide who owns the file.
        f = h5py.File(fn, 'a')
        f.close()

    def _setup_data_io(self):
        if getattr(self, "io", None) is not None: return
        self.io = io_registry[self.dataset_type](self.dataset)

    @parallel_root_only
    def save_data(self, array, node, name, set_attr=None, force=False, passthrough = False):
        """
        Arbitrary numpy data will be saved to the region in the datafile
        described by *node* and *name*.  If data file does not exist, it throws
        no error and simply does not save.
        """

        if self._data_mode != 'a': return
        try:
            node_loc = self._data_file[node]
            if name in node_loc and force:
                mylog.info("Overwriting node %s/%s", node, name)
                del self._data_file[node][name]
            elif name in node_loc and passthrough:
                return
        except Exception: pass
        myGroup = self._data_file['/']
        for q in node.split('/'):
            if q: myGroup = myGroup.require_group(q)
        arr = myGroup.create_dataset(name,data=array)
        if set_attr is not None:
            for i, j in set_attr.items(): arr.attrs[i] = j
        self._data_file.flush()

    def _reload_data_file(self, *args, **kwargs):
        if self._data_file is None: return
        self._data_file.close()
        del self._data_file
        self._data_file = h5py.File(self.__data_filename, self._data_mode)

    def save_object(self, obj, name):
        """
        Save an object (*obj*) to the data_file using the Pickle protocol,
        under the name *name* on the node /Objects.
        """
        s = cPickle.dumps(obj, protocol=-1)
        self.save_data(np.array(s, dtype='c'), "/Objects", name, force = True)

    def load_object(self, name):
        """
        Load and return and object from the data_file using the Pickle protocol,
        under the name *name* on the node /Objects.
        """
        obj = self.get_data("/Objects", name)
        if obj is None:
            return
        obj = cPickle.loads(obj.value)
        if iterable(obj) and len(obj) == 2:
            obj = obj[1] # Just the object, not the ds
        if hasattr(obj, '_fix_pickle'): obj._fix_pickle()
        return obj

    def get_data(self, node, name):
        """
        Return the dataset with a given *name* located at *node* in the
        datafile.
        """
        if self._data_file is None:
            return None
        if node[0] != "/": node = "/%s" % node

        myGroup = self._data_file['/']
        for group in node.split('/'):
            if group:
                if group not in myGroup:
                    return None
                myGroup = myGroup[group]
        if name not in myGroup:
            return None

        full_name = "%s/%s" % (node, name)
        try:
            return self._data_file[full_name][:]
        except TypeError:
            return self._data_file[full_name]

    def _get_particle_type_counts(self):
        # this is implemented by subclasses
        raise NotImplementedError

    def _close_data_file(self):
        if self._data_file:
            self._data_file.close()
            del self._data_file
            self._data_file = None

    def _split_fields(self, fields):
        # This will split fields into either generated or read fields
        fields_to_read, fields_to_generate = [], []
        for ftype, fname in fields:
            if fname in self.field_list or (ftype, fname) in self.field_list:
                fields_to_read.append((ftype, fname))
            elif fname in self.ds.derived_field_list or (ftype, fname) in self.ds.derived_field_list:
                fields_to_generate.append((ftype, fname))
            else:
                raise YTFieldNotFound((ftype,fname), self.ds)
        return fields_to_read, fields_to_generate

    def _read_particle_fields(self, fields, dobj, chunk = None):
        if len(fields) == 0: return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        if chunk is None:
            self._identify_base_chunk(dobj)
        fields_to_return = self.io._read_particle_selection(
            self._chunk_io(dobj, cache = False),
            selector,
            fields_to_read)
        return fields_to_return, fields_to_generate

    def _read_fluid_fields(self, fields, dobj, chunk = None):
        if len(fields) == 0: return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        if chunk is None:
            self._identify_base_chunk(dobj)
            chunk_size = dobj.size
        else:
            chunk_size = chunk.data_size
        fields_to_return = self.io._read_fluid_selection(
            self._chunk_io(dobj),
            selector,
            fields_to_read,
            chunk_size)
        return fields_to_return, fields_to_generate

    def _chunk(self, dobj, chunking_style, ngz = 0, **kwargs):
        # A chunk is either None or (grids, size)
        if dobj._current_chunk is None:
            self._identify_base_chunk(dobj)
        if ngz != 0 and chunking_style != "spatial":
            raise NotImplementedError
        if chunking_style == "all":
            return self._chunk_all(dobj, **kwargs)
        elif chunking_style == "spatial":
            return self._chunk_spatial(dobj, ngz, **kwargs)
        elif chunking_style == "io":
            return self._chunk_io(dobj, **kwargs)
        else:
            raise NotImplementedError

def cached_property(func):
    n = '_%s' % func.__name__
    def cached_func(self):
        if self._cache and getattr(self, n, None) is not None:
            return getattr(self, n)
        if self.data_size is None:
            tr = self._accumulate_values(n[1:])
        else:
            tr = func(self)
        if self._cache:

            setattr(self, n, tr)
        return tr
    return property(cached_func)

class YTDataChunk(object):

    def __init__(self, dobj, chunk_type, objs, data_size = None,
                 field_type = None, cache = False, fast_index = None):
        self.dobj = dobj
        self.chunk_type = chunk_type
        self.objs = objs
        self.data_size = data_size
        self._field_type = field_type
        self._cache = cache
        self._fast_index = fast_index

    def _accumulate_values(self, method):
        # We call this generically.  It's somewhat slower, since we're doing
        # costly getattr functions, but this allows us to generalize.
        mname = "select_%s" % method
        arrs = []
        for obj in self._fast_index or self.objs:
            f = getattr(obj, mname)
            arrs.append(f(self.dobj))
        if method == "dtcoords":
            arrs = [arr[0] for arr in arrs]
        elif method == "tcoords":
            arrs = [arr[1] for arr in arrs]
        arrs = uconcatenate(arrs)
        self.data_size = arrs.shape[0]
        return arrs

    @cached_property
    def fcoords(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_fcoords(
                self.dobj.selector, self.data_size)
            ci = YTArray(ci, input_units = "code_length",
                         registry = self.dobj.ds.unit_registry)
            return ci
        ci = np.empty((self.data_size, 3), dtype='float64')
        ci = YTArray(ci, input_units = "code_length",
                     registry = self.dobj.ds.unit_registry)
        if self.data_size == 0: return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_fcoords(self.dobj)
            if c.shape[0] == 0: continue
            ci[ind:ind+c.shape[0], :] = c
            ind += c.shape[0]
        return ci

    @cached_property
    def icoords(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_icoords(
                self.dobj.selector, self.data_size)
            return ci
        ci = np.empty((self.data_size, 3), dtype='int64')
        if self.data_size == 0: return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_icoords(self.dobj)
            if c.shape[0] == 0: continue
            ci[ind:ind+c.shape[0], :] = c
            ind += c.shape[0]
        return ci

    @cached_property
    def fwidth(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_fwidth(
                self.dobj.selector, self.data_size)
            ci = YTArray(ci, input_units = "code_length",
                         registry = self.dobj.ds.unit_registry)
            return ci
        ci = np.empty((self.data_size, 3), dtype='float64')
        ci = YTArray(ci, input_units = "code_length",
                     registry = self.dobj.ds.unit_registry)
        if self.data_size == 0: return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_fwidth(self.dobj)
            if c.shape[0] == 0: continue
            ci[ind:ind+c.shape[0], :] = c
            ind += c.shape[0]
        return ci

    @cached_property
    def ires(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_ires(
                self.dobj.selector, self.data_size)
            return ci
        ci = np.empty(self.data_size, dtype='int64')
        if self.data_size == 0: return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_ires(self.dobj)
            if c.shape == 0: continue
            ci[ind:ind+c.size] = c
            ind += c.size
        return ci

    @cached_property
    def tcoords(self):
        self.dtcoords
        return self._tcoords

    @cached_property
    def dtcoords(self):
        ct = np.empty(self.data_size, dtype='float64')
        cdt = np.empty(self.data_size, dtype='float64')
        self._tcoords = ct # Se this for tcoords
        if self.data_size == 0: return cdt
        ind = 0
        for obj in self._fast_index or self.objs:
            gdt, gt = obj.select_tcoords(self.dobj)
            if gt.size == 0: continue
            ct[ind:ind+gt.size] = gt
            cdt[ind:ind+gdt.size] = gdt
            ind += gt.size
        return cdt

    @cached_property
    def fcoords_vertex(self):
        nodes_per_elem = self.dobj.index.meshes[0].connectivity_indices.shape[1]
        dim = self.dobj.ds.dimensionality
        ci = np.empty((self.data_size, nodes_per_elem, dim), dtype='float64')
        ci = YTArray(ci, input_units = "code_length",
                     registry = self.dobj.ds.unit_registry)
        if self.data_size == 0: return ci
        ind = 0
        for obj in self.objs:
            c = obj.select_fcoords_vertex(self.dobj)
            if c.shape[0] == 0: continue
            ci[ind:ind+c.shape[0], :, :] = c
            ind += c.shape[0]
        return ci


class ChunkDataCache(object):
    def __init__(self, base_iter, preload_fields, geometry_handler,
                 max_length = 256):
        # At some point, max_length should instead become a heuristic function,
        # potentially looking at estimated memory usage.  Note that this never
        # initializes the iterator; it assumes the iterator is already created,
        # and it calls next() on it.
        self.base_iter = base_iter.__iter__()
        self.queue = []
        self.max_length = max_length
        self.preload_fields = preload_fields
        self.geometry_handler = geometry_handler
        self.cache = {}

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if len(self.queue) == 0:
            for i in range(self.max_length):
                try:
                    self.queue.append(next(self.base_iter))
                except StopIteration:
                    break
            # If it's still zero ...
            if len(self.queue) == 0: raise StopIteration
            chunk = YTDataChunk(None, "cache", self.queue, cache=False)
            self.cache = self.geometry_handler.io._read_chunk_data(
                chunk, self.preload_fields) or {}
        g = self.queue.pop(0)
        g._initialize_cache(self.cache.pop(g.id, {}))
        return g

def is_curvilinear(geo):
    # tell geometry is curvilinear or not
    if geo in ["polar", "cylindrical", "spherical"]:
        return True
    else:
        return False


"""
yt is a toolkit for analyzing and visualizing volumetric data.

* Website: https://yt-project.org
* Documentation: https://yt-project.org/doc
* Data hub: https://girder.hub.yt
* Contribute: https://github.com/yt-project/yt

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

__version__ = "3.6.0"

# First module imports
import numpy as np # For modern purposes
import numpy # In case anyone wishes to use it by name

from yt.funcs import \
    iterable, \
    get_memory_usage, \
    print_tb, \
    rootonly, \
    insert_ipython, \
    get_pbar, \
    only_on_root, \
    is_root, \
    get_version_stack, \
    get_yt_supp, \
    get_yt_version, \
    parallel_profile, \
    enable_plugins, \
    memory_checker, \
    deprecated_class, \
    toggle_interactivity
from yt.utilities.logger import ytLogger as mylog

import yt.utilities.physical_constants as physical_constants
import yt.units as units
from yt.units.unit_object import define_unit
from yt.units.yt_array import \
    YTArray, \
    YTQuantity, \
    uconcatenate, \
    ucross, \
    uintersect1d, \
    uunion1d, \
    unorm, \
    udot, \
    ustack, \
    uvstack, \
    uhstack, \
    loadtxt, \
    savetxt, \
    display_ytarray

from yt.fields.api import \
    field_plugins, \
    DerivedField, \
    FieldDetector, \
    FieldInfoContainer, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType, \
    add_field, \
    derived_field, \
    add_xray_emissivity_field

from yt.data_objects.api import \
    DatasetSeries, ImageArray, \
    particle_filter, add_particle_filter, \
    create_profile, Profile1D, Profile2D, Profile3D, \
    ParticleProfile

# For backwards compatibility
TimeSeriesData = deprecated_class(DatasetSeries)

from yt.frontends.api import _frontend_container
frontends = _frontend_container()

from yt.frontends.stream.api import \
    load_uniform_grid, load_amr_grids, \
    load_particles, load_hexahedral_mesh, load_octree, \
    hexahedral_connectivity, load_unstructured_mesh

from yt.frontends.ytdata.api import \
    save_as_dataset

# For backwards compatibility
GadgetDataset = frontends.gadget.GadgetDataset
GadgetStaticOutput = deprecated_class(GadgetDataset)
TipsyDataset = frontends.tipsy.TipsyDataset
TipsyStaticOutput = deprecated_class(TipsyDataset)

# Now individual component imports from the visualization API
from yt.visualization.api import \
    FixedResolutionBuffer, ObliqueFixedResolutionBuffer, \
    write_bitmap, write_image, \
    apply_colormap, scale_image, write_projection, \
    SlicePlot, AxisAlignedSlicePlot, OffAxisSlicePlot, LinePlot, \
    LineBuffer, ProjectionPlot, OffAxisProjectionPlot, \
    show_colormaps, add_colormap, make_colormap, \
    ProfilePlot, PhasePlot, ParticlePhasePlot, \
    ParticleProjectionPlot, ParticleImageBuffer, ParticlePlot, \
    FITSImageData, FITSSlice, FITSProjection, FITSOffAxisSlice, \
    FITSOffAxisProjection, plot_2d

from yt.visualization.volume_rendering.api import \
    volume_render, create_scene, ColorTransferFunction, TransferFunction, \
    off_axis_projection, interactive_render
import yt.visualization.volume_rendering.api as volume_rendering
#    TransferFunctionHelper, MultiVariateTransferFunction
#    off_axis_projection

from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects, enable_parallelism, communication_system

from yt.convenience import \
    load, simulation

from yt.testing import run_nose

# Import some helpful math utilities
from yt.utilities.math_utils import \
    ortho_find, quartiles, periodic_position

from yt.units.unit_systems import UnitSystem
from yt.units.unit_object import unit_system_registry

from yt.analysis_modules.list_modules import \
    amods



import numpy as np
from yt.testing import \
    fake_random_ds, \
    assert_equal, \
    assert_rel_equal, \
    fake_amr_ds
from yt.units.unit_object import Unit
import os
import tempfile

LENGTH_UNIT = 2.0


def setup():
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


def teardown_func(fns):
    for fn in fns:
        try:
            os.remove(fn)
        except OSError:
            pass

def test_projection():
    fns = []
    for nprocs in [8, 1]:
        # We want to test both 1 proc and 8 procs, to make sure that
        # parallelism isn't broken
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ('g/cm**3', 'K', 'cm/s', 'cm/s', 'cm/s')
        ds = fake_random_ds(64, fields=fields, units=units, nprocs=nprocs,
                            length_unit=LENGTH_UNIT)
        dims = ds.domain_dimensions
        xn, yn, zn = ds.domain_dimensions
        xi, yi, zi = ds.domain_left_edge.to_ndarray() + \
            1.0 / (ds.domain_dimensions * 2)
        xf, yf, zf = ds.domain_right_edge.to_ndarray() - \
            1.0 / (ds.domain_dimensions * 2)
        dd = ds.all_data()
        coords = np.mgrid[xi:xf:xn*1j, yi:yf:yn*1j, zi:zf:zn*1j]
        uc = [np.unique(c) for c in coords]
        # test if projections inherit the field parameters of their data sources
        dd.set_field_parameter("bulk_velocity", np.array([0,1,2]))
        proj = ds.proj("density", 0, data_source=dd)
        assert_equal(dd.field_parameters["bulk_velocity"],
                     proj.field_parameters["bulk_velocity"])

        # Some simple projection tests with single grids
        for ax, an in enumerate("xyz"):
            xax = ds.coordinates.x_axis[ax]
            yax = ds.coordinates.y_axis[ax]
            for wf in ['density', ("gas", "density"), None]:
                proj = ds.proj(["ones", "density"], ax, weight_field=wf)
                if wf is None:
                    assert_equal(proj["ones"].sum(),
                                 LENGTH_UNIT*proj["ones"].size)
                    assert_equal(proj["ones"].min(), LENGTH_UNIT)
                    assert_equal(proj["ones"].max(), LENGTH_UNIT)
                else:
                    assert_equal(proj["ones"].sum(), proj["ones"].size)
                    assert_equal(proj["ones"].min(), 1.0)
                    assert_equal(proj["ones"].max(), 1.0)
                assert_equal(np.unique(proj["px"]), uc[xax])
                assert_equal(np.unique(proj["py"]), uc[yax])
                assert_equal(np.unique(proj["pdx"]), 1.0/(dims[xax]*2.0))
                assert_equal(np.unique(proj["pdy"]), 1.0/(dims[yax]*2.0))
                plots = [proj.to_pw(fields='density'), proj.to_pw()]
                for pw in plots:
                    for p in pw.plots.values():
                        tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
                        os.close(tmpfd)
                        p.save(name=tmpname)
                        fns.append(tmpname)
                frb = proj.to_frb((1.0, 'unitary'), 64)
                for proj_field in ['ones', 'density', 'temperature']:
                    fi = ds._get_field_info(proj_field)
                    assert_equal(frb[proj_field].info['data_source'],
                                 proj.__str__())
                    assert_equal(frb[proj_field].info['axis'], ax)
                    assert_equal(frb[proj_field].info['field'], proj_field)
                    field_unit = Unit(fi.units)
                    if wf is not None:
                        assert_equal(
                            frb[proj_field].units,
                            Unit(field_unit, registry=ds.unit_registry))
                    else:
                        if frb[proj_field].units.is_code_unit:
                            proj_unit = "code_length"
                        else:
                            proj_unit = "cm"
                        if field_unit != '' and field_unit != Unit():
                            proj_unit = \
                                "({0}) * {1}".format(field_unit, proj_unit)
                        assert_equal(
                            frb[proj_field].units,
                            Unit(proj_unit, registry=ds.unit_registry))
                    assert_equal(frb[proj_field].info['xlim'], frb.bounds[:2])
                    assert_equal(frb[proj_field].info['ylim'], frb.bounds[2:])
                    assert_equal(frb[proj_field].info['center'], proj.center)
                    if wf is None:
                        assert_equal(frb[proj_field].info['weight_field'], wf)
                    else:
                        assert_equal(
                            frb[proj_field].info['weight_field'],
                            proj.data_source._determine_fields(wf)[0])
            # wf == None
            assert_equal(wf, None)
            v1 = proj["density"].sum()
            v2 = (dd["density"] * dd["d%s" % an]).sum()
            assert_rel_equal(v1, v2.in_units(v1.units), 10)
    teardown_func(fns)


def test_max_level():
    ds = fake_amr_ds()
    proj = ds.proj('Density', 2, method='mip', max_level=2)
    assert proj['grid_level'].max() == 2

    proj = ds.proj('Density', 2, method='mip')
    assert proj['grid_level'].max() == ds.index.max_level



from yt.testing import fake_particle_ds, requires_module
import tempfile

@requires_module('firefly_api')
def test_firefly_JSON_object():
    tmpdir = tempfile.mkdtemp()

    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        path_to_firefly = tmpdir,
        velocity_units = 'cm/s',
        coordinate_units = 'cm',
        dataset_name='test')
    reader.dumpToJSON()




"""
OWLS frontend tests using the snapshot_033 dataset




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import OrderedDict

from yt.testing import \
    requires_file
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    data_dir_load, \
    sph_answer
from yt.frontends.owls.api import OWLSDataset
from yt.data_objects.particle_filters import add_particle_filter

os33 = "snapshot_033/snap_033.0.hdf5"

# This maps from field names to weight field names to use for projections
_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ("gas", "density")),
        (('gas', 'He_p0_number_density'), None),
        (('gas', 'velocity_magnitude'), None),
        (("deposit", "all_density"), None),
        (("deposit", "all_count"), None),
        (("deposit", "all_cic"), None),
        (("deposit", "PartType0_density"), None),
        (("deposit", "PartType4_density"), None),
    ]
)


@requires_ds(os33, big_data=True)
def test_snapshot_033():
    ds = data_dir_load(os33)
    for test in sph_answer(ds, 'snap_033', 2*128**3, _fields):
        test_snapshot_033.__name__ = test.description
        yield test


@requires_file(os33)
def test_OWLSDataset():
    assert isinstance(data_dir_load(os33), OWLSDataset)
    
    
@requires_ds(os33)
def test_OWLS_particlefilter():
    ds = data_dir_load(os33)
    ad = ds.all_data()
    def cold_gas(pfilter, data):
        temperature = data[pfilter.filtered_type, "Temperature"]
        filter = (temperature.in_units('K') <= 1e5)
        return filter
    add_particle_filter("gas_cold", function=cold_gas, filtered_type='PartType0', requires=["Temperature"])
    ds.add_particle_filter('gas_cold')

    mask = (ad['PartType0','Temperature'] <= 1e5)
    assert ad['PartType0','Temperature'][mask].shape == ad['gas_cold','Temperature'].shape




import numpy as np

from yt import \
    load
from yt.testing import \
    fake_random_ds, \
    assert_equal, \
    assert_rel_equal, \
    requires_file
from yt.units.yt_array import \
    uconcatenate

def test_ray():
    for nproc in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=nproc)
        dx = (ds.domain_right_edge - ds.domain_left_edge) / \
          ds.domain_dimensions
        # Three we choose, to get varying vectors, and ten random
        pp1 = np.random.random((3, 13))
        pp2 = np.random.random((3, 13))
        pp1[:,0] = [0.1, 0.2, 0.3]
        pp2[:,0] = [0.8, 0.1, 0.4]
        pp1[:,1] = [0.9, 0.2, 0.3]
        pp2[:,1] = [0.8, 0.1, 0.4]
        pp1[:,2] = [0.9, 0.2, 0.9]
        pp2[:,2] = [0.8, 0.1, 0.4]
        unitary = ds.arr(1.0, '')
        for i in range(pp1.shape[1]):
            p1 = ds.arr(pp1[:,i] + 1e-8 * np.random.random(3), 'code_length')
            p2 = ds.arr(pp2[:,i] + 1e-8 * np.random.random(3), 'code_length')

            my_ray = ds.ray(p1, p2)
            assert_rel_equal(my_ray['dts'].sum(), unitary, 14)
            ray_cells = my_ray['dts'] > 0

            # find cells intersected by the ray
            my_all = ds.all_data()
            
            dt = np.abs(dx / (p2 - p1))
            tin  = uconcatenate([[(my_all['x'] - p1[0]) / (p2 - p1)[0] - 0.5 * dt[0]],
                                 [(my_all['y'] - p1[1]) / (p2 - p1)[1] - 0.5 * dt[1]],
                                 [(my_all['z'] - p1[2]) / (p2 - p1)[2] - 0.5 * dt[2]]])
            tout = uconcatenate([[(my_all['x'] - p1[0]) / (p2 - p1)[0] + 0.5 * dt[0]],
                                 [(my_all['y'] - p1[1]) / (p2 - p1)[1] + 0.5 * dt[1]],
                                 [(my_all['z'] - p1[2]) / (p2 - p1)[2] + 0.5 * dt[2]]])
            tin = tin.max(axis=0)
            tout = tout.min(axis=0)
            my_cells = (tin < tout) & (tin < 1) & (tout > 0)

            assert_equal(ray_cells.sum(), my_cells.sum())
            assert_rel_equal(my_ray['density'][ray_cells].sum(),
                             my_all['density'][my_cells].sum(), 14)
            assert_rel_equal(my_ray['dts'].sum(), unitary, 14)

@requires_file('GadgetDiskGalaxy/snapshot_200.hdf5')
def test_ray_in_particle_octree():
    ds = load('GadgetDiskGalaxy/snapshot_200.hdf5')
    start = ds.arr([31995.63476562, 31473.6640625, 28969.88671875], "code_length")
    end = ds.arr([31995.63476562, 31473.6640625, 29219.88671875], "code_length")
    ray = ds.ray(start, end)
    ray["t"]
    assert_equal(ray["dts"].sum(dtype="f8"), 1.0)



"""
GadgetFOF frontend tests using gadget_fof datasets



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.frontends.gadget_fof.api import \
    GadgetFOFDataset
from yt.testing import \
    requires_file, \
    assert_equal, \
    assert_array_equal
from yt.utilities.answer_testing.framework import \
    FieldValuesTest, \
    requires_ds, \
    data_dir_load

_fields = ("particle_position_x", "particle_position_y",
           "particle_position_z", "particle_velocity_x",
           "particle_velocity_y", "particle_velocity_z",
           "particle_mass", "particle_identifier")

# a dataset with empty files
g5 = "gadget_fof_halos/groups_005/fof_subhalo_tab_005.0.hdf5"
g42 = "gadget_fof_halos/groups_042/fof_subhalo_tab_042.0.hdf5"


@requires_ds(g5)
def test_fields_g5():
    for field in _fields:
        yield FieldValuesTest(g5, field, particle_type=True)


@requires_ds(g42)
def test_fields_g42():
    for field in _fields:
        yield FieldValuesTest(g42, field, particle_type=True)

@requires_file(g42)
def test_GadgetFOFDataset():
    assert isinstance(data_dir_load(g42), GadgetFOFDataset)

# fof/subhalo catalog with member particles
g298 = "gadget_halos/data/groups_298/fof_subhalo_tab_298.0.hdf5"

@requires_file(g298)
def test_subhalos():
    ds = data_dir_load(g298)
    total_sub = 0
    total_int = 0
    for hid in range(0, ds.index.particle_count["Group"]):
        my_h = ds.halo("Group", hid)
        h_ids = my_h["ID"]
        for sid in range(int(my_h["subhalo_number"][0])):
            my_s = ds.halo("Subhalo", (my_h.particle_identifier, sid))
            total_sub += my_s["ID"].size
            total_int += np.intersect1d(h_ids, my_s["ID"]).size

    # Test that all subhalo particles are contained within
    # their parent group.
    assert_equal(total_sub, total_int)

@requires_file(g298)
def test_halo_masses():
    ds = data_dir_load(g298)
    ad = ds.all_data()
    for ptype in ["Group", "Subhalo"]:
        nhalos = ds.index.particle_count[ptype]
        mass = ds.arr(np.zeros(nhalos), "code_mass")
        for i in range(nhalos):
            halo = ds.halo(ptype, i)
            mass[i] = halo.mass

        # Check that masses from halo containers are the same
        # as the array of all masses.  This will test getting
        # scalar fields for halos correctly.
        assert_array_equal(ad[ptype, "particle_mass"], mass)

# fof/subhalo catalog with no member ids in first file
g56 = "gadget_halos/data/groups_056/fof_subhalo_tab_056.0.hdf5"

# This dataset has halos in one file and ids in another,
# which can confuse the field detection.
@requires_file(g56)
def test_unbalanced_dataset():
    ds = data_dir_load(g56)
    halo = ds.halo("Group", 0)
    halo["member_ids"]
    assert True

# fof/subhalo catalog with no member ids in first file
g76 = "gadget_halos/data/groups_076/fof_subhalo_tab_076.0.hdf5"

# This dataset has one halo with particles distributed over 3 files
# with the 2nd file being empty.
@requires_file(g76)
def test_3file_halo():
    ds = data_dir_load(g76)
    # this halo's particles are distributed over 3 files with the
    # middle file being empty
    halo = ds.halo("Group", 6)
    halo["member_ids"]
    assert True



"""
FITS-specific data structures
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import stat
import numpy as np
import numpy.core.defchararray as np_char
import os
import re
import time
import uuid
import weakref
import warnings

from collections import defaultdict

from yt.config import ytcfg
from yt.funcs import \
    ensure_list, \
    mylog, \
    setdefaultattr
from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.geometry.grid_geometry_handler import \
    GridIndex
from yt.geometry.geometry_handler import \
    YTDataChunk
from yt.data_objects.static_output import \
    Dataset
from yt.utilities.file_handler import \
    FITSFileHandler
from yt.utilities.io_handler import \
    io_registry
from .fields import FITSFieldInfo, \
    WCSFITSFieldInfo
from yt.utilities.decompose import \
    decompose_array, get_psize
from yt.funcs import issue_deprecation_warning
from yt.units.unit_lookup_table import \
    default_unit_symbol_lut, \
    prefixable_units, \
    unit_prefixes
from yt.units import dimensions
from yt.utilities.on_demand_imports import \
    _astropy, NotAModule

lon_prefixes = ["X","RA","GLON","LINEAR"]
lat_prefixes = ["Y","DEC","GLAT","LINEAR"]
delimiters = ["*", "/", "-", "^", "(", ")"]
delimiters += [str(i) for i in range(10)]
regex_pattern = '|'.join(re.escape(_) for _ in delimiters)

spec_names = {"V": "Velocity",
              "F": "Frequency",
              "E": "Energy",
              "W": "Wavelength"}

space_prefixes = list(set(lon_prefixes + lat_prefixes))
sky_prefixes = set(space_prefixes)
sky_prefixes.difference_update({"X", "Y", "LINEAR"})
sky_prefixes = list(sky_prefixes)
spec_prefixes = list(spec_names.keys())

field_from_unit = {"Jy":"intensity",
                   "K":"temperature"}

class FITSGrid(AMRGridPatch):
    _id_offset = 0
    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename,
                              index=index)
        self.Parent = None
        self.Children = []
        self.Level = 0

    def __repr__(self):
        return "FITSGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

class FITSHierarchy(GridIndex):

    grid = FITSGrid

    def __init__(self,ds,dataset_type='fits'):
        self.dataset_type = dataset_type
        self.field_indexes = {}
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self._handle = ds._handle
        self.float_type = np.float64
        GridIndex.__init__(self,ds,dataset_type)

    def _initialize_data_storage(self):
        pass

    def _guess_name_from_units(self, units):
        for k,v in field_from_unit.items():
            if k in units:
                mylog.warning("Guessing this is a %s field based on its units of %s." % (v,k))
                return v
        return None

    def _determine_image_units(self, header, known_units):
        try:
            field_units = header["bunit"].lower().strip(" ").replace(" ", "")
            # FITS units always return upper-case, so we need to get
            # the right case by comparing against known units. This
            # only really works for common units.
            units = set(re.split(regex_pattern, field_units))
            if '' in units: 
                units.remove('')
            n = int(0)
            for unit in units:
                if unit in known_units:
                    field_units = field_units.replace(unit, known_units[unit])
                    n += 1
            if n != len(units) or n == 0:
                field_units = "dimensionless"
            if field_units[0] == "/":
                field_units = "1%s" % field_units
            return field_units
        except KeyError:
            return "dimensionless"

    def _ensure_same_dims(self, hdu):
        ds = self.dataset
        conditions = [hdu.header["naxis"] != ds.primary_header["naxis"]]
        for i in range(ds.naxis):
            nax = "naxis%d" % (i+1)
            conditions.append(hdu.header[nax] != ds.primary_header[nax])
        if np.any(conditions):
            return False
        else:
            return True

    def _detect_output_fields(self):
        self.field_list = []
        self._axis_map = {}
        self._file_map = {}
        self._ext_map = {}
        self._scale_map = {}
        dup_field_index = {}
        # Since FITS header keywords are case-insensitive, we only pick a subset of
        # prefixes, ones that we expect to end up in headers.
        known_units = dict(
            [(unit.lower(), unit) for unit in self.ds.unit_registry.lut]
        )
        for unit in list(known_units.values()):
            if unit in prefixable_units:
                for p in ["n","u","m","c","k"]:
                    known_units[(p+unit).lower()] = p+unit
        # We create a field from each slice on the 4th axis
        if self.dataset.naxis == 4:
            naxis4 = self.dataset.primary_header["naxis4"]
        else:
            naxis4 = 1
        for i, fits_file in enumerate(self.dataset._handle._fits_files):
            for j, hdu in enumerate(fits_file):
                if isinstance(hdu, _astropy.pyfits.BinTableHDU) or hdu.header["naxis"] == 0:
                    continue
                if self._ensure_same_dims(hdu):
                    units = self._determine_image_units(hdu.header, known_units)
                    try:
                        # Grab field name from btype
                        fname = hdu.header["btype"]
                    except KeyError:
                        # Try to guess the name from the units
                        fname = self._guess_name_from_units(units)
                        # When all else fails
                        if fname is None: fname = "image_%d" % (j)
                    if self.ds.num_files > 1 and fname.startswith("image"):
                        fname += "_file_%d" % (i)
                    if ("fits", fname) in self.field_list:
                        if fname in dup_field_index:
                            dup_field_index[fname] += 1
                        else:
                            dup_field_index[fname] = 1
                        mylog.warning("This field has the same name as a previously loaded " +
                                      "field. Changing the name from %s to %s_%d. To avoid " %
                                      (fname, fname, dup_field_index[fname]) +
                                      " this, change one of the BTYPE header keywords.")
                        fname += "_%d" % (dup_field_index[fname])
                    for k in range(naxis4):
                        if naxis4 > 1:
                            fname += "_%s_%d" % (hdu.header["CTYPE4"], k+1)
                        self._axis_map[fname] = k
                        self._file_map[fname] = fits_file
                        self._ext_map[fname] = j
                        self._scale_map[fname] = [0.0, 1.0]
                        if "bzero" in hdu.header:
                            self._scale_map[fname][0] = hdu.header["bzero"]
                        if "bscale" in hdu.header:
                            self._scale_map[fname][1] = hdu.header["bscale"]
                        self.field_list.append(("fits", fname))
                        self.dataset.field_units[fname] = units
                        mylog.info("Adding field %s to the list of fields." % (fname))
                        if units == "dimensionless":
                            mylog.warning("Could not determine dimensions for field %s, " % (fname) +
                                          "setting to dimensionless.")
                else:
                    mylog.warning("Image block %s does not have " % (hdu.name.lower()) +
                                  "the same dimensions as the primary and will not be " +
                                  "available as a field.")

    def _count_grids(self):
        self.num_grids = self.ds.parameters["nprocs"]

    def _parse_index(self):
        ds = self.dataset

        # If nprocs > 1, decompose the domain into virtual grids
        if self.num_grids > 1:
            self._domain_decomp()
        else:
            self.grid_left_edge[0,:] = ds.domain_left_edge
            self.grid_right_edge[0,:] = ds.domain_right_edge
            self.grid_dimensions[0] = ds.domain_dimensions

        self.grid_levels.flat[:] = 0
        self.grids = np.empty(self.num_grids, dtype='object')
        for i in range(self.num_grids):
            self.grids[i] = self.grid(i, self, self.grid_levels[i,0])

    def _domain_decomp(self):
        bbox = np.array([self.ds.domain_left_edge,
                         self.ds.domain_right_edge]).transpose()
        dims = self.ds.domain_dimensions
        psize = get_psize(dims, self.num_grids)
        gle, gre, shapes, slices = decompose_array(dims, psize, bbox)
        self.grid_left_edge = self.ds.arr(gle, "code_length")
        self.grid_right_edge = self.ds.arr(gre, "code_length")
        self.grid_dimensions = np.array(shapes, dtype="int32")

    def _populate_grid_objects(self):
        for i in range(self.num_grids):
            self.grids[i]._prepare_grid()
            self.grids[i]._setup_dx()
        self.max_level = 0

    def _setup_derived_fields(self):
        super(FITSHierarchy, self)._setup_derived_fields()
        [self.dataset.conversion_factors[field]
         for field in self.field_list]
        for field in self.field_list:
            if field not in self.derived_field_list:
                self.derived_field_list.append(field)

        for field in self.derived_field_list:
            f = self.dataset.field_info[field]
            if f._function.__name__ == "_TranslationFunc":
                # Translating an already-converted field
                self.dataset.conversion_factors[field] = 1.0

    def _setup_data_io(self):
        self.io = io_registry[self.dataset_type](self.dataset)

    def _chunk_io(self, dobj, cache = True, local_only = False):
        # local_only is only useful for inline datasets and requires
        # implementation by subclasses.
        gfiles = defaultdict(list)
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for g in gobjs:
            gfiles[g.id].append(g)
        for fn in sorted(gfiles):
            gs = gfiles[fn]
            yield YTDataChunk(dobj, "io", gs, self._count_selection(dobj, gs),
                              cache = cache)

def find_primary_header(fileh):
    # Sometimes the primary hdu doesn't have an image
    if len(fileh) > 1 and fileh[0].header["naxis"] == 0:
        first_image = 1
    else:
        first_image = 0
    header = fileh[first_image].header
    return header, first_image

def check_fits_valid(args):
    ext = args[0].rsplit(".", 1)[-1]
    if ext.upper() in ("GZ", "FZ"):
        # We don't know for sure that there will be > 1
        ext = args[0].rsplit(".", 1)[0].rsplit(".", 1)[-1]
    if ext.upper() not in ("FITS", "FTS"):
        return None
    elif isinstance(_astropy.pyfits, NotAModule):
        raise RuntimeError("This appears to be a FITS file, but AstroPy is not installed.")
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning, append=True)
            fileh = _astropy.pyfits.open(args[0])
        header, _ = find_primary_header(fileh)
        if header["naxis"] >= 2:
            return fileh
        else:
            fileh.close()
    except Exception: pass
    return None

def check_sky_coords(args, ndim):
    fileh = check_fits_valid(args)
    if fileh is not None:
        try:
            if (len(fileh) > 1 and
                fileh[1].name == "EVENTS" and ndim == 2):
                fileh.close()
                return True
            else:
                header, _ = find_primary_header(fileh)
                if header["naxis"] < ndim:
                    return False
                axis_names = [header.get("ctype%d" % (i + 1), "")
                              for i in range(header["naxis"])]
                x = find_axes(axis_names, sky_prefixes + spec_prefixes)
                fileh.close()
                return x >= ndim
        except Exception: pass
    return False

class FITSDataset(Dataset):
    _index_class = FITSHierarchy
    _field_info_class = FITSFieldInfo
    _dataset_type = "fits"
    _handle = None

    def __init__(self, filename,
                 dataset_type='fits',
                 auxiliary_files=[],
                 nprocs=None,
                 storage_filename=None,
                 nan_mask=None,
                 suppress_astropy_warnings=True,
                 parameters=None,
                 units_override=None,
                 unit_system="cgs"):

        if parameters is None:
            parameters = {}
        parameters["nprocs"] = nprocs
        self.specified_parameters = parameters

        if suppress_astropy_warnings:
            warnings.filterwarnings('ignore', module="astropy", append=True)
        auxiliary_files = ensure_list(auxiliary_files)
        self.filenames = [filename] + auxiliary_files
        self.num_files = len(self.filenames)
        self.fluid_types += ("fits",)
        if nan_mask is None:
            self.nan_mask = {}
        elif isinstance(nan_mask, float):
            self.nan_mask = {"all":nan_mask}
        elif isinstance(nan_mask, dict):
            self.nan_mask = nan_mask
        self._handle = FITSFileHandler(self.filenames[0])
        if (isinstance(self.filenames[0], _astropy.pyfits.hdu.image._ImageBaseHDU) or
            isinstance(self.filenames[0], _astropy.pyfits.HDUList)):
            fn = "InMemoryFITSFile_%s" % uuid.uuid4().hex
        else:
            fn = self.filenames[0]
        self._handle._fits_files.append(self._handle)
        if self.num_files > 1:
            for fits_file in auxiliary_files:
                if isinstance(fits_file, _astropy.pyfits.hdu.image._ImageBaseHDU):
                    f = _astropy.pyfits.HDUList([fits_file])
                elif isinstance(fits_file, _astropy.pyfits.HDUList):
                    f = fits_file
                else:
                    if os.path.exists(fits_file):
                        fn = fits_file
                    else:
                        fn = os.path.join(ytcfg.get("yt","test_data_dir"),fits_file)
                    f = _astropy.pyfits.open(fn, memmap=True,
                                             do_not_scale_image_data=True,
                                             ignore_blank=True)
                self._handle._fits_files.append(f)

        self.refine_by = 2

        Dataset.__init__(self, fn, dataset_type, units_override=units_override,
                         unit_system=unit_system)
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        default_length_units = [u for u,v in default_unit_symbol_lut.items()
                                if str(v[1]) == "(length)"]
        more_length_units = []
        for unit in default_length_units:
            if unit in prefixable_units:
                more_length_units += [prefix+unit for prefix in unit_prefixes]
        default_length_units += more_length_units
        file_units = []
        cunits = [self.wcs.wcs.cunit[i] for i in range(self.dimensionality)]
        for unit in (_.to_string() for _ in cunits):
            if unit in default_length_units:
                file_units.append(unit)
        if len(set(file_units)) == 1:
            length_factor = self.wcs.wcs.cdelt[0]
            length_unit = str(file_units[0])
            mylog.info("Found length units of %s." % (length_unit))
        else:
            self.no_cgs_equiv_length = True
            mylog.warning("No length conversion provided. Assuming 1 = 1 cm.")
            length_factor = 1.0
            length_unit = "cm"
        setdefaultattr(self, 'length_unit', self.quan(length_factor,length_unit))
        setdefaultattr(self, 'mass_unit', self.quan(1.0, "g"))
        setdefaultattr(self, 'time_unit', self.quan(1.0, "s"))
        setdefaultattr(self, 'velocity_unit', self.quan(1.0, "cm/s"))
        if "beam_size" in self.specified_parameters:
            beam_size = self.specified_parameters["beam_size"]
            beam_size = self.quan(beam_size[0], beam_size[1]).in_cgs().value
        else:
            beam_size = 1.0
        self.unit_registry.add("beam", beam_size, dimensions=dimensions.solid_angle)

    def _parse_parameter_file(self):

        self._determine_structure()
        self._determine_axes()

        if self.parameter_filename.startswith("InMemory"):
            self.unique_identifier = time.time()
        else:
            self.unique_identifier = \
                int(os.stat(self.parameter_filename)[stat.ST_CTIME])

        # Determine dimensionality

        self.dimensionality = self.naxis
        self.geometry = "cartesian"

        # Sometimes a FITS file has a 4D datacube, in which case
        # we take the 4th axis and assume it consists of different fields.
        if self.dimensionality == 4: 
            self.dimensionality = 3

        self._determine_wcs()

        self.domain_dimensions = np.array(self.dims)[:self.dimensionality]
        if self.dimensionality == 2:
            self.domain_dimensions = np.append(self.domain_dimensions,
                                               [int(1)])

        domain_left_edge = np.array([0.5]*3)
        domain_right_edge = np.array([float(dim)+0.5 for dim in self.domain_dimensions])

        if self.dimensionality == 2:
            domain_left_edge[-1] = 0.5
            domain_right_edge[-1] = 1.5

        self.domain_left_edge = domain_left_edge
        self.domain_right_edge = domain_right_edge

        # Get the simulation time
        try:
            self.current_time = self.parameters["time"]
        except Exception:
            mylog.warning("Cannot find time")
            self.current_time = 0.0
            pass

        # For now we'll ignore these
        self.periodicity = (False,)*3
        self.current_redshift = self.omega_lambda = self.omega_matter = \
            self.hubble_constant = self.cosmological_simulation = 0.0

        self._determine_nprocs()

        # Now we can set up some of our parameters for convenience.
        for k, v in self.primary_header.items():
            self.parameters[k] = v
        # Remove potential default keys
        self.parameters.pop('', None)

    def _determine_nprocs(self):
        # If nprocs is None, do some automatic decomposition of the domain
        if self.specified_parameters["nprocs"] is None:
            nprocs = np.around(np.prod(self.domain_dimensions)/32**self.dimensionality).astype("int")
            self.parameters["nprocs"] = max(min(nprocs, 512), 1)
        else:
            self.parameters["nprocs"] = self.specified_parameters["nprocs"]

    def _determine_structure(self):
        self.primary_header, self.first_image = find_primary_header(self._handle)
        self.naxis = self.primary_header["naxis"]
        self.axis_names = [self.primary_header.get("ctype%d" % (i+1),"LINEAR")
                           for i in range(self.naxis)]
        self.dims = [self.primary_header["naxis%d" % (i+1)]
                     for i in range(self.naxis)]

    def _determine_wcs(self):
        wcs = _astropy.pywcs.WCS(header=self.primary_header)
        if self.naxis == 4:
            self.wcs = _astropy.pywcs.WCS(naxis=3)
            self.wcs.wcs.crpix = wcs.wcs.crpix[:3]
            self.wcs.wcs.cdelt = wcs.wcs.cdelt[:3]
            self.wcs.wcs.crval = wcs.wcs.crval[:3]
            self.wcs.wcs.cunit = [str(unit) for unit in wcs.wcs.cunit][:3]
            self.wcs.wcs.ctype = [type for type in wcs.wcs.ctype][:3]
        else:
            self.wcs = wcs

    def _determine_axes(self):
        self.lat_axis = 1
        self.lon_axis = 0
        self.lat_name = "Y"
        self.lon_name = "X"

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        fileh = check_fits_valid(args)
        if fileh is None:
            return False
        else:
            fileh.close()
            return True

    @classmethod
    def _guess_candidates(cls, base, directories, files):
        candidates = []
        for fn, fnl in ((_, _.lower()) for _ in files):
            if fnl.endswith(".fits") or \
               fnl.endswith(".fits.gz") or \
               fnl.endswith(".fits.fz"):
                candidates.append(fn)
        # FITS files don't preclude subdirectories
        return candidates, True

    def close(self):
        self._handle.close()

def find_axes(axis_names, prefixes):
    x = 0
    for p in prefixes:
        y = np_char.startswith(axis_names, p)
        x += np.any(y)
    return x

class SkyDataFITSDataset(FITSDataset):
    _field_info_class = WCSFITSFieldInfo

    def _determine_wcs(self):
        super(SkyDataFITSDataset, self)._determine_wcs()
        end = min(self.dimensionality+1, 4)
        self.ctypes = np.array([self.primary_header["CTYPE%d" % (i)]
                                for i in range(1, end)])
        self.wcs_2d = self.wcs

    def _parse_parameter_file(self):
        super(SkyDataFITSDataset, self)._parse_parameter_file()

        end = min(self.dimensionality + 1, 4)

        self.geometry = "spectral_cube"

        log_str = "Detected these axes: "+"%s "*len(self.ctypes)
        mylog.info(log_str % tuple([ctype for ctype in self.ctypes]))

        self.lat_axis = np.zeros((end-1), dtype="bool")
        for p in lat_prefixes:
            self.lat_axis += np_char.startswith(self.ctypes, p)
        self.lat_axis = np.where(self.lat_axis)[0][0]
        self.lat_name = self.ctypes[self.lat_axis].split("-")[0].lower()

        self.lon_axis = np.zeros((end-1), dtype="bool")
        for p in lon_prefixes:
            self.lon_axis += np_char.startswith(self.ctypes, p)
        self.lon_axis = np.where(self.lon_axis)[0][0]
        self.lon_name = self.ctypes[self.lon_axis].split("-")[0].lower()

        if self.lat_axis == self.lon_axis and self.lat_name == self.lon_name:
            self.lat_axis = 1
            self.lon_axis = 0
            self.lat_name = "Y"
            self.lon_name = "X"

        self.spec_axis = 2
        self.spec_name = "z"
        self.spec_unit = ""

    def _set_code_unit_attributes(self):
        super(SkyDataFITSDataset, self)._set_code_unit_attributes()
        units = self.wcs_2d.wcs.cunit[0]
        if units == "deg":
            units = "degree"
        if units == "rad":
            units = "radian"
        pixel_area = np.prod(np.abs(self.wcs_2d.wcs.cdelt))
        pixel_area = self.quan(pixel_area, "%s**2" % (units)).in_cgs()
        pixel_dims = pixel_area.units.dimensions
        self.unit_registry.add("pixel", float(pixel_area.value), dimensions=pixel_dims)

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        return check_sky_coords(args, 2)

class SpectralCubeFITSHierarchy(FITSHierarchy):

    def _domain_decomp(self):
        dz = self.ds.quan(1.0, "code_length") * self.ds.spectral_factor
        self.grid_dimensions[:, 2] = np.around(float(self.ds.domain_dimensions[2]) /
                                               self.num_grids).astype("int")
        self.grid_dimensions[-1, 2] += (self.ds.domain_dimensions[2] % self.num_grids)
        self.grid_left_edge[0, 2] = self.ds.domain_left_edge[2]
        self.grid_left_edge[1:, 2] = self.ds.domain_left_edge[2] + \
                                     np.cumsum(self.grid_dimensions[:-1, 2]) * dz
        self.grid_right_edge[:, 2] = self.grid_left_edge[:, 2] + self.grid_dimensions[:, 2] * dz
        self.grid_left_edge[:, :2] = self.ds.domain_left_edge[:2]
        self.grid_right_edge[:, :2] = self.ds.domain_right_edge[:2]
        self.grid_dimensions[:, :2] = self.ds.domain_dimensions[:2]


class SpectralCubeFITSDataset(SkyDataFITSDataset):
    _index_class = SpectralCubeFITSHierarchy
    def __init__(self, filename,
                 auxiliary_files=[],
                 nprocs=None,
                 storage_filename=None,
                 nan_mask=None,
                 spectral_factor=1.0,
                 suppress_astropy_warnings=True,
                 parameters=None,
                 units_override=None,
                 unit_system="cgs",
                 z_axis_decomp=None):
        self.spectral_factor = spectral_factor
        if z_axis_decomp is not None:
            issue_deprecation_warning("The 'z_axis_decomp' argument is deprecated, "
                                      "as this decomposition is now performed for "
                                      "spectral-cube FITS datasets automatically.")
        super(SpectralCubeFITSDataset, self).__init__(filename, nprocs=nprocs,
            auxiliary_files=auxiliary_files, storage_filename=storage_filename,
            suppress_astropy_warnings=suppress_astropy_warnings, nan_mask=nan_mask,
            parameters=parameters, units_override=units_override,
            unit_system=unit_system)

    def _parse_parameter_file(self):
        super(SpectralCubeFITSDataset, self)._parse_parameter_file()

        self.geometry = "spectral_cube"

        end = min(self.dimensionality+1,4)

        self.spec_axis = np.zeros(end-1, dtype="bool")
        for p in spec_names.keys():
            self.spec_axis += np_char.startswith(self.ctypes, p)
        self.spec_axis = np.where(self.spec_axis)[0][0]
        self.spec_name = spec_names[self.ctypes[self.spec_axis].split("-")[0][0]]

        # Extract a subimage from a WCS object
        self.wcs_2d = self.wcs.sub(["longitude", "latitude"])

        self._p0 = self.wcs.wcs.crpix[self.spec_axis]
        self._dz = self.wcs.wcs.cdelt[self.spec_axis]
        self._z0 = self.wcs.wcs.crval[self.spec_axis]
        self.spec_unit = str(self.wcs.wcs.cunit[self.spec_axis])

        if self.spectral_factor == "auto":
            self.spectral_factor = float(max(self.domain_dimensions[[self.lon_axis,
                                                                     self.lat_axis]]))
            self.spectral_factor /= self.domain_dimensions[self.spec_axis]
            mylog.info("Setting the spectral factor to %f" % (self.spectral_factor))
        Dz = self.domain_right_edge[self.spec_axis]-self.domain_left_edge[self.spec_axis]
        dre = self.domain_right_edge
        dre[self.spec_axis] = (self.domain_left_edge[self.spec_axis] +
                               self.spectral_factor*Dz)
        self.domain_right_edge = dre
        self._dz /= self.spectral_factor
        self._p0 = (self._p0-0.5)*self.spectral_factor + 0.5

    def _determine_nprocs(self):
        # If nprocs is None, do some automatic decomposition of the domain
        if self.specified_parameters["nprocs"] is None:
            nprocs = np.around(self.domain_dimensions[2] / 8).astype("int")
            self.parameters["nprocs"] = max(min(nprocs, 512), 1)
        else:
            self.parameters["nprocs"] = self.specified_parameters["nprocs"]

    def spec2pixel(self, spec_value):
        sv = self.arr(spec_value).in_units(self.spec_unit)
        return self.arr((sv.v-self._z0)/self._dz+self._p0,
                        "code_length")

    def pixel2spec(self, pixel_value):
        pv = self.arr(pixel_value, "code_length")
        return self.arr((pv.v-self._p0)*self._dz+self._z0,
                        self.spec_unit)

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        return check_sky_coords(args, 3)

class EventsFITSHierarchy(FITSHierarchy):

    def _detect_output_fields(self):
        ds = self.dataset
        self.field_list = []
        for k,v in ds.events_info.items():
            fname = "event_"+k
            mylog.info("Adding field %s to the list of fields." % (fname))
            self.field_list.append(("io",fname))
            if k in ["x","y"]:
                field_unit = "code_length"
            else:
                field_unit = v
            self.dataset.field_units[("io",fname)] = field_unit
        return

    def _parse_index(self):
        super(EventsFITSHierarchy, self)._parse_index()
        try:
            self.grid_particle_count[:] = self.dataset.primary_header["naxis2"]
        except KeyError:
            self.grid_particle_count[:] = 0.0
        self._particle_indices = np.zeros(self.num_grids + 1, dtype='int64')
        self._particle_indices[1] = self.grid_particle_count.squeeze()

class EventsFITSDataset(SkyDataFITSDataset):
    _index_class = EventsFITSHierarchy
    def __init__(self, filename,
                 storage_filename=None,
                 suppress_astropy_warnings=True,
                 reblock=1,
                 parameters=None,
                 units_override=None,
                 unit_system="cgs"):
        self.reblock = reblock
        super(EventsFITSDataset, self).__init__(filename, nprocs=1,
            storage_filename=storage_filename, parameters=parameters,
            suppress_astropy_warnings=suppress_astropy_warnings,
            units_override=units_override, unit_system=unit_system)

    def _determine_structure(self):
        self.first_image = 1
        self.primary_header = self._handle[self.first_image].header
        self.naxis = 2

    def _determine_wcs(self):
        self.wcs = _astropy.pywcs.WCS(naxis=2)
        self.events_info = {}
        for k, v in self.primary_header.items():
            if k.startswith("TTYP"):
                if v.lower() in ["x", "y"]:
                    num = k.strip("TTYPE")
                    self.events_info[v.lower()] = (self.primary_header["TLMIN" + num],
                                                   self.primary_header["TLMAX" + num],
                                                   self.primary_header["TCTYP" + num],
                                                   self.primary_header["TCRVL" + num],
                                                   self.primary_header["TCDLT" + num],
                                                   self.primary_header["TCRPX" + num])
                elif v.lower() in ["energy", "time"]:
                    num = k.strip("TTYPE")
                    unit = self.primary_header["TUNIT" + num].lower()
                    if unit.endswith("ev"):
                        unit = unit.replace("ev", "eV")
                    self.events_info[v.lower()] = unit
        self.axis_names = [self.events_info[ax][2] for ax in ["x", "y"]]
        if "reblock" in self.specified_parameters:
            issue_deprecation_warning("'reblock' is now a keyword argument that "
                                      "can be passed to 'yt.load'. This behavior "
                                      "is deprecated.")
            self.reblock = self.specified_parameters["reblock"]
        self.wcs.wcs.cdelt = [self.events_info["x"][4] * self.reblock,
                              self.events_info["y"][4] * self.reblock]
        self.wcs.wcs.crpix = [(self.events_info["x"][5] - 0.5) / self.reblock + 0.5,
                              (self.events_info["y"][5] - 0.5) / self.reblock + 0.5]
        self.wcs.wcs.ctype = [self.events_info["x"][2], self.events_info["y"][2]]
        self.wcs.wcs.cunit = ["deg", "deg"]
        self.wcs.wcs.crval = [self.events_info["x"][3], self.events_info["y"][3]]
        self.dims = [(self.events_info["x"][1] - self.events_info["x"][0]) / self.reblock,
                     (self.events_info["y"][1] - self.events_info["y"][0]) / self.reblock]
        self.ctypes = self.axis_names
        self.wcs_2d = self.wcs

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        fileh = check_fits_valid(args)
        if fileh is not None:
            try:
                valid = fileh[1].name == "EVENTS"
                fileh.close()
                return valid
            except Exception: pass
        return False



"""
Particle operations for Lagrangian Volume



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os, sys
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects

from yt.config import ytcfg

cdef import from "particle.h":
    struct particle:
        np.int64_t id
        float pos[6]

ctypedef struct particleflat:
    np.int64_t id
    float pos_x
    float pos_y
    float pos_z
    float vel_x
    float vel_y
    float vel_z

cdef import from "halo.h":
    struct halo:
        np.int64_t id
        float pos[6]
        float corevel[3]
        float bulkvel[3]
        float J[3]
        float m, r, child_r, mgrav, vmax, rvmax, rs, vrms, energy, spin
        np.int64_t num_p, num_child_particles, p_start, desc, flags, n_core
        float min_pos_err, min_vel_err, min_bulkvel_err

cdef import from "io_generic.h":
    ctypedef void (*LPG) (char *filename, particle **p, np.int64_t *num_p)
    ctypedef void (*AHG) (halo *h, particle *hp)
    void set_load_particles_generic(LPG func, AHG afunc)

cdef import from "rockstar.h":
    void rockstar(float *bounds, np.int64_t manual_subs)

cdef import from "config.h":
    void setup_config()

cdef import from "server.h" nogil:
    int server()
    np.int64_t READER_TYPE
    np.int64_t WRITER_TYPE

cdef import from "client.h" nogil:
    void client(np.int64_t in_type)

cdef import from "meta_io.h":
    void read_particles(char *filename)
    void output_halos(np.int64_t id_offset, np.int64_t snap,
                      np.int64_t chunk, float *bounds)

cdef import from "config_vars.h":
    # Rockstar cleverly puts all of the config variables inside a templated
    # definition of their variables.
    char *FILE_FORMAT
    np.float64_t PARTICLE_MASS

    char *MASS_DEFINITION
    np.int64_t MIN_HALO_OUTPUT_SIZE
    np.float64_t FORCE_RES

    np.float64_t SCALE_NOW
    np.float64_t h0
    np.float64_t Ol
    np.float64_t Om

    np.int64_t GADGET_ID_BYTES
    np.float64_t GADGET_MASS_CONVERSION
    np.float64_t GADGET_LENGTH_CONVERSION
    np.int64_t GADGET_SKIP_NON_HALO_PARTICLES
    np.int64_t RESCALE_PARTICLE_MASS

    np.int64_t PARALLEL_IO
    char *PARALLEL_IO_SERVER_ADDRESS
    char *PARALLEL_IO_SERVER_PORT
    np.int64_t PARALLEL_IO_WRITER_PORT
    char *PARALLEL_IO_SERVER_INTERFACE
    char *RUN_ON_SUCCESS

    char *INBASE
    char *FILENAME
    np.int64_t STARTING_SNAP
    np.int64_t RESTART_SNAP
    np.int64_t NUM_SNAPS
    np.int64_t NUM_BLOCKS
    np.int64_t NUM_READERS
    np.int64_t PRELOAD_PARTICLES
    char *SNAPSHOT_NAMES
    char *LIGHTCONE_ALT_SNAPS
    char *BLOCK_NAMES

    char *OUTBASE
    np.float64_t OVERLAP_LENGTH
    np.int64_t NUM_WRITERS
    np.int64_t FORK_READERS_FROM_WRITERS
    np.int64_t FORK_PROCESSORS_PER_MACHINE

    char *OUTPUT_FORMAT
    np.int64_t DELETE_BINARY_OUTPUT_AFTER_FINISHED
    np.int64_t FULL_PARTICLE_CHUNKS
    char *BGC2_SNAPNAMES

    np.int64_t BOUND_PROPS
    np.int64_t BOUND_OUT_TO_HALO_EDGE
    np.int64_t DO_MERGER_TREE_ONLY
    np.int64_t IGNORE_PARTICLE_IDS
    np.float64_t TRIM_OVERLAP
    np.float64_t ROUND_AFTER_TRIM
    np.int64_t LIGHTCONE
    np.int64_t PERIODIC

    np.float64_t LIGHTCONE_ORIGIN[3]
    np.float64_t LIGHTCONE_ALT_ORIGIN[3]

    np.float64_t LIMIT_CENTER[3]
    np.float64_t LIMIT_RADIUS

    np.int64_t SWAP_ENDIANNESS
    np.int64_t GADGET_VARIANT

    np.float64_t FOF_FRACTION
    np.float64_t FOF_LINKING_LENGTH
    np.float64_t INCLUDE_HOST_POTENTIAL_RATIO
    np.float64_t DOUBLE_COUNT_SUBHALO_MASS_RATIO
    np.int64_t TEMPORAL_HALO_FINDING
    np.int64_t MIN_HALO_PARTICLES
    np.float64_t UNBOUND_THRESHOLD
    np.int64_t ALT_NFW_METRIC

    np.int64_t TOTAL_PARTICLES
    np.float64_t BOX_SIZE
    np.int64_t OUTPUT_HMAD
    np.int64_t OUTPUT_PARTICLES
    np.int64_t OUTPUT_LEVELS
    np.float64_t DUMP_PARTICLES[3]

    np.float64_t AVG_PARTICLE_SPACING
    np.int64_t SINGLE_SNAP

# Forward declare
cdef class RockstarInterface

cdef void rh_analyze_halo(halo *h, particle *hp):
    # I don't know why, but sometimes we get halos with 0 particles.
    if h.num_p == 0: return
    cdef particleflat[:] pslice
    pslice = <particleflat[:h.num_p]> (<particleflat *>hp)
    parray = np.asarray(pslice)
    for cb in rh.callbacks:
        cb(rh.ds, parray)
    # This is where we call our functions

cdef void rh_read_particles(char *filename, particle **p, np.int64_t *num_p):
    global SCALE_NOW
    cdef np.float64_t left_edge[6]
    cdef np.ndarray[np.int64_t, ndim=1] arri
    cdef np.ndarray[np.float64_t, ndim=1] arr
    cdef unsigned long long pi,fi,i
    cdef np.int64_t local_parts = 0
    ds = rh.ds = next(rh.tsl)

    SCALE_NOW = 1.0/(ds.current_redshift+1.0)
    # First we need to find out how many this reader is going to read in
    # if the number of readers > 1.
    dd = ds.all_data()

    # Add particle type filter if not defined
    if rh.particle_type not in ds.particle_types and rh.particle_type != 'all':
        ds.add_particle_filter(rh.particle_type)

    if NUM_BLOCKS > 1:
        local_parts = 0
        for chunk in parallel_objects(
                dd.chunks([], "io")):
            local_parts += chunk[rh.particle_type, "particle_ones"].sum()
    else:
        local_parts = TOTAL_PARTICLES

    p[0] = <particle *> malloc(sizeof(particle) * local_parts)

    left_edge[0] = ds.domain_left_edge.in_units('Mpccm/h')[0]
    left_edge[1] = ds.domain_left_edge.in_units('Mpccm/h')[1]
    left_edge[2] = ds.domain_left_edge.in_units('Mpccm/h')[2]
    left_edge[3] = left_edge[4] = left_edge[5] = 0.0
    pi = 0
    for chunk in parallel_objects(dd.chunks([], "io")):
        arri = np.asarray(chunk[rh.particle_type, "particle_index"],
                          dtype="int64")
        npart = arri.size
        for i in range(npart):
            p[0][i+pi].id = <np.int64_t> arri[i]
        fi = 0
        for field in ["particle_position_x", "particle_position_y",
                      "particle_position_z",
                      "particle_velocity_x", "particle_velocity_y",
                      "particle_velocity_z"]:
            if "position" in field:
                unit = "Mpccm/h"
            else:
                unit = "km/s"
            arr = chunk[rh.particle_type, field].in_units(unit).astype("float64")
            for i in range(npart):
                p[0][i+pi].pos[fi] = (arr[i]-left_edge[fi])
            fi += 1
        pi += npart
    num_p[0] = local_parts
    del ds

cdef class RockstarInterface:

    cdef public object data_source
    cdef public object ts
    cdef public object tsl
    cdef public object ds
    cdef int rank
    cdef int size
    cdef public int block_ratio
    cdef public object particle_type
    cdef public int total_particles
    cdef public object callbacks

    def __cinit__(self, ts):
        self.ts = ts
        self.tsl = ts.__iter__() #timeseries generator used by read

    def setup_rockstar(self, char *server_address, char *server_port,
                       int num_snaps, np.int64_t total_particles,
                       particle_type,
                       np.float64_t particle_mass,
                       int parallel = False, int num_readers = 1,
                       int num_writers = 1,
                       int writing_port = -1, int block_ratio = 1,
                       int periodic = 1, force_res=None,
                       int min_halo_size = 25, outbase = "None",
                       callbacks = None, int restart_num = 0):
        global PARALLEL_IO, PARALLEL_IO_SERVER_ADDRESS, PARALLEL_IO_SERVER_PORT
        global FILENAME, FILE_FORMAT, NUM_SNAPS, STARTING_SNAP, h0, Ol, Om
        global BOX_SIZE, PERIODIC, PARTICLE_MASS, NUM_BLOCKS, NUM_READERS
        global FORK_READERS_FROM_WRITERS, PARALLEL_IO_WRITER_PORT, NUM_WRITERS
        global rh, SCALE_NOW, OUTBASE, MIN_HALO_OUTPUT_SIZE, OUTPUT_FORMAT
        global OVERLAP_LENGTH, TOTAL_PARTICLES, FORCE_RES, RESTART_SNAP
        if force_res is not None:
            FORCE_RES=np.float64(force_res)
            #print "set force res to ",FORCE_RES
        OVERLAP_LENGTH = 0.0
        if parallel:
            PARALLEL_IO = 1
            PARALLEL_IO_SERVER_ADDRESS = server_address
            PARALLEL_IO_SERVER_PORT = server_port
            if writing_port > 0:
                PARALLEL_IO_WRITER_PORT = writing_port
        else:
            PARALLEL_IO = 0
            PARALLEL_IO_SERVER_ADDRESS = server_address
            PARALLEL_IO_SERVER_PORT = server_port
        FILENAME = "inline.<block>"
        FILE_FORMAT = "GENERIC"
        OUTPUT_FORMAT = "ASCII"
        NUM_SNAPS = num_snaps
        RESTART_SNAP = restart_num
        NUM_READERS = num_readers
        NUM_WRITERS = num_writers
        NUM_BLOCKS = num_readers
        MIN_HALO_OUTPUT_SIZE=min_halo_size
        TOTAL_PARTICLES = total_particles
        self.block_ratio = block_ratio
        self.particle_type = particle_type

        tds = self.ts[0]
        h0 = tds.hubble_constant
        Ol = tds.omega_lambda
        Om = tds.omega_matter
        SCALE_NOW = 1.0/(tds.current_redshift+1.0)
        if callbacks is None: callbacks = []
        self.callbacks = callbacks
        if not outbase == 'None'.encode('UTF-8'):
            #output directory. since we can't change the output filenames
            #workaround is to make a new directory
            OUTBASE = outbase

        PARTICLE_MASS = particle_mass
        PERIODIC = periodic
        BOX_SIZE = (tds.domain_right_edge[0] -
                    tds.domain_left_edge[0]).in_units("Mpccm/h")
        setup_config()
        rh = self
        cdef LPG func = rh_read_particles
        cdef AHG afunc = rh_analyze_halo
        set_load_particles_generic(func, afunc)

    def call_rockstar(self):
        read_particles("generic")
        rockstar(NULL, 0)
        output_halos(0, 0, 0, NULL)

    def start_server(self):
        with nogil:
            server()

    def start_reader(self):
        cdef np.int64_t in_type = np.int64(READER_TYPE)
        client(in_type)

    def start_writer(self):
        cdef np.int64_t in_type = np.int64(WRITER_TYPE)
        client(in_type)




import glob
import os
from yt.extern.six import add_metaclass
from yt.convenience import \
    load
from yt.funcs import \
    only_on_root
from yt.utilities.exceptions import \
    YTOutputNotIdentified
from yt.utilities.logger import ytLogger as \
    mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects
from yt.data_objects.time_series import \
    DatasetSeries, \
    RegisteredSimulationTimeSeries


@add_metaclass(RegisteredSimulationTimeSeries)
class ExodusIISimulation(DatasetSeries):
    r"""
    Initialize an ExodusII Simulation object.

    Upon creation, the input directory is searched for valid ExodusIIDatasets.
    The get_time_series can be used to generate a DatasetSeries object.

    simulation_directory : str
        The directory that contain the simulation data.
    
    Examples
    --------
    >>> import yt
    >>> sim = yt.simulation("demo_second", "ExodusII")
    >>> sim.get_time_series()
    >>> for ds in sim:
    ...     print ds.current_time

    """
    
    def __init__(self, simulation_directory, find_outputs=False):
        self.simulation_directory = simulation_directory
        fn_pattern = "%s/*" % self.simulation_directory
        potential_outputs = glob.glob(fn_pattern)
        self.all_outputs = self._check_for_outputs(potential_outputs)
        self.all_outputs.sort(key=lambda obj: obj["filename"])

    def __iter__(self):
        for o in self._pre_outputs:
            fn, step = o
            ds = load(fn, step=step)
            self._setup_function(ds)
            yield ds

    def __getitem__(self, key):
        if isinstance(key, slice):
            if isinstance(key.start, float):
                return self.get_range(key.start, key.stop)
            # This will return a sliced up object!
            return DatasetSeries(self._pre_outputs[key], self.parallel)
        o = self._pre_outputs[key]
        fn, step = o
        o = load(fn, step=step)
        self._setup_function(o)
        return o

    def get_time_series(self, parallel=False, setup_function=None):
        r"""
        Instantiate a DatasetSeries object for a set of outputs.

        If no additional keywords given, a DatasetSeries object will be
        created with all potential datasets created by the simulation.

        Fine-level filtering is currently not implemented.
        
        """

        all_outputs = self.all_outputs
        ds_list = []
        for output in all_outputs:
            num_steps = output['num_steps']
            fn = output['filename']
            for step in range(num_steps):
                ds_list.append((fn, step))
        super(ExodusIISimulation, self).__init__(ds_list, 
                                                 parallel=parallel, 
                                                 setup_function=setup_function)
        
    def _check_for_outputs(self, potential_outputs):
        r"""
        Check a list of files to see if they are valid datasets.
        """

        only_on_root(mylog.info, "Checking %d potential outputs.",
                     len(potential_outputs))

        my_outputs = {}
        for my_storage, output in parallel_objects(potential_outputs,
                                                   storage=my_outputs):
            if os.path.exists(output):
                try:
                    ds = load(output)
                    if ds is not None:
                        num_steps = ds.num_steps
                        my_storage.result = {"filename": output,
                                             "num_steps": num_steps}
                except YTOutputNotIdentified:
                    mylog.error("Failed to load %s", output)
        my_outputs = [my_output for my_output in my_outputs.values() \
                      if my_output is not None]
        return my_outputs



"""
This is where we define a handful of vector operations for fields.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from .derived_field import \
    ValidateParameter, \
    ValidateSpatial

from yt.utilities.math_utils import \
    get_sph_r_component, \
    get_sph_theta_component, \
    get_sph_phi_component, \
    get_cyl_r_component, \
    get_cyl_z_component, \
    get_cyl_theta_component

from yt.funcs import \
    just_one, \
    handle_mks_cgs

from yt.geometry.geometry_handler import \
    is_curvilinear

from yt.utilities.lib.misc_utilities import obtain_relative_velocity_vector

def get_bulk(data, basename, unit):
    if data.has_field_parameter("bulk_%s" % basename):
        bulk = data.get_field_parameter("bulk_%s" % basename)
    else:
        bulk = [0, 0, 0]*unit
    return bulk


def create_magnitude_field(registry, basename, field_units,
                           ftype="gas", slice_info=None,
                           validators=None, particle_type=False):

    axis_order = registry.ds.coordinates.axis_order

    field_components = [(ftype, "%s_%s" % (basename, ax)) for ax in axis_order]

    def _magnitude(field, data):
        fn = field_components[0]
        if data.has_field_parameter('bulk_%s' % basename):
            fn = (fn[0], 'relative_%s' % fn[1])
        d = data[fn]
        mag = (d)**2
        for idim in range(1, registry.ds.dimensionality):
            fn = field_components[idim]
            if data.has_field_parameter('bulk_%s' % basename):
                fn = (fn[0], 'relative_%s' % fn[1])
            mag += (data[fn])**2
        return np.sqrt(mag)

    if particle_type is True:
        sampling_type = 'particle'
    else:
        sampling_type = 'cell'

    registry.add_field((ftype, "%s_magnitude" % basename),
                       sampling_type=sampling_type,
                       function=_magnitude,
                       units=field_units,
                       validators=validators)

def create_relative_field(registry, basename, field_units, ftype='gas',
                          slice_info=None, validators=None,
                          particle_type=False):

    axis_order = registry.ds.coordinates.axis_order

    field_components = [(ftype, "%s_%s" % (basename, ax)) for ax in axis_order]

    def relative_vector(ax):
        def _relative_vector(field, data):
            iax = axis_order.index(ax)
            d = handle_mks_cgs(data[field_components[iax]], field.units)
            bulk = handle_mks_cgs(get_bulk(data, basename, d.unit_quantity),
                                  field.units)
            return d - bulk[iax]
        return _relative_vector

    if particle_type is True:
        sampling_type = 'particle'
    else:
        sampling_type = 'cell'

    for d in axis_order:
        registry.add_field((ftype, "relative_%s_%s" % (basename, d)),
                           sampling_type=sampling_type,
                           function=relative_vector(d),
                           units=field_units,
                           validators=validators)


def create_squared_field(registry, basename, field_units,
                         ftype="gas", slice_info=None,
                         validators=None, particle_type=False):

    axis_order = registry.ds.coordinates.axis_order

    field_components = [(ftype, "%s_%s" % (basename, ax)) for ax in axis_order]

    def _squared(field, data):
        fn = field_components[0]
        if data.has_field_parameter('bulk_%s' % basename):
            fn = (fn[0], 'relative_%s' % fn[1])
        squared = data[fn] * data[fn]
        for idim in range(1, registry.ds.dimensionality):
            fn = field_components[idim]
            squared += data[fn] * data[fn]
        return squared

    registry.add_field((ftype, "%s_squared" % basename), sampling_type="cell",
                       function=_squared, units=field_units,
                       validators=validators, particle_type=particle_type)

def create_vector_fields(registry, basename, field_units,
                         ftype="gas", slice_info=None):
    from yt.units.unit_object import Unit
    # slice_info would be the left, the right, and the factor.
    # For example, with the old Enzo-ZEUS fields, this would be:
    # slice(None, -2, None)
    # slice(1, -1, None)
    # 1.0
    # Otherwise, we default to a centered difference.
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info

    xn, yn, zn = [(ftype, "%s_%s" % (basename, ax)) for ax in 'xyz']

    # Is this safe?
    if registry.ds.dimensionality < 3:
        zn = ("index", "zeros")
    if registry.ds.dimensionality < 2:
        yn = ("index", "zeros")

    create_relative_field(
        registry, basename, field_units, ftype=ftype, slice_info=slice_info,
        validators=[ValidateParameter('bulk_%s' % basename)])

    create_magnitude_field(
        registry, basename, field_units, ftype=ftype, slice_info=slice_info,
        validators=[ValidateParameter('bulk_%s' % basename)])

    if not is_curvilinear(registry.ds.geometry):

        # The following fields are invalid for curvilinear geometries
        def _spherical_radius_component(field, data):
            """The spherical radius component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            theta = data['index', 'spherical_theta']
            phi = data['index', 'spherical_phi']
            rv = get_sph_r_component(vectors, theta, phi, normal)
            # Now, anywhere that radius is in fact zero, we want to zero out our
            # return values.
            rv[np.isnan(theta)] = 0.0
            return rv

        registry.add_field((ftype, "%s_spherical_radius" % basename),
                           sampling_type="cell",
                           function=_spherical_radius_component,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center"),
                                       ValidateParameter("bulk_%s" % basename)])

        def _radial(field, data):
            return data[ftype, "%s_spherical_radius" % basename]

        def _radial_absolute(field, data):
            return np.abs(data[ftype, "%s_spherical_radius" % basename])

        def _tangential(field, data):
            return np.sqrt(data[ftype, "%s_spherical_theta" % basename]**2.0 +
                           data[ftype, "%s_spherical_phi" % basename]**2.0)

        registry.add_field((ftype, "radial_%s" % basename),
                           sampling_type="cell",
                           function=_radial,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center")])

        registry.add_field((ftype, "radial_%s_absolute" % basename),
                           sampling_type="cell",
                           function=_radial_absolute,
                           units=field_units)

        registry.add_field((ftype, "tangential_%s" % basename),
                           sampling_type="cell",
                           function=_tangential,
                           units=field_units)

        def _spherical_theta_component(field, data):
            """The spherical theta component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            theta = data["index", "spherical_theta"]
            phi = data["index", "spherical_phi"]
            return get_sph_theta_component(vectors, theta, phi, normal)

        registry.add_field((ftype, "%s_spherical_theta" % basename),
                           sampling_type="cell",
                           function=_spherical_theta_component,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center"),
                                       ValidateParameter("bulk_%s" % basename)])

        def _spherical_phi_component(field, data):
            """The spherical phi component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            phi = data["index", "spherical_phi"]
            return get_sph_phi_component(vectors, phi, normal)

        registry.add_field((ftype, "%s_spherical_phi" % basename),
                           sampling_type="cell",
                           function=_spherical_phi_component,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center"),
                                       ValidateParameter("bulk_%s" % basename)])

        def _cp_vectors(ax):
            def _cp_val(field, data):
                vec = data.get_field_parameter("cp_%s_vec" % (ax))
                tr  = data[xn[0], 'relative_%s' % xn[1]] * vec.d[0]
                tr += data[yn[0], 'relative_%s' % yn[1]] * vec.d[1]
                tr += data[zn[0], 'relative_%s' % zn[1]] * vec.d[2]
                return tr
            return _cp_val

        for ax in 'xyz':
            registry.add_field((ftype, "cutting_plane_%s_%s" % (basename, ax)),
                               sampling_type="cell",
                               function=_cp_vectors(ax),
                               units=field_units)

        def _divergence(field, data):
            ds = div_fac * just_one(data["index", "dx"])
            f  = data[xn[0], 'relative_%s' % xn[1]][sl_right,1:-1,1:-1]/ds
            f -= data[xn[0], 'relative_%s' % xn[1]][sl_left ,1:-1,1:-1]/ds
            ds = div_fac * just_one(data["index", "dy"])
            f += data[yn[0], 'relative_%s' % yn[1]][1:-1,sl_right,1:-1]/ds
            f -= data[yn[0], 'relative_%s' % yn[1]][1:-1,sl_left ,1:-1]/ds
            ds = div_fac * just_one(data["index", "dz"])
            f += data[zn[0], 'relative_%s' % zn[1]][1:-1,1:-1,sl_right]/ds
            f -= data[zn[0], 'relative_%s' % zn[1]][1:-1,1:-1,sl_left ]/ds
            new_field = data.ds.arr(
                np.zeros(data[xn].shape, dtype=np.float64), f.units)
            new_field[1:-1,1:-1,1:-1] = f
            return new_field

        def _divergence_abs(field, data):
            return np.abs(data[ftype, "%s_divergence" % basename])

        field_units = Unit(field_units, registry=registry.ds.unit_registry)
        div_units = field_units / registry.ds.unit_system["length"]

        registry.add_field((ftype, "%s_divergence" % basename),
                           sampling_type="cell",
                           function=_divergence,
                           units=div_units,
                           validators=[ValidateSpatial(1),
                                       ValidateParameter('bulk_%s' % basename)])

        registry.add_field((ftype, "%s_divergence_absolute" % basename),
                           sampling_type="cell",
                           function=_divergence_abs,
                           units=div_units)

        def _tangential_over_magnitude(field, data):
            tr = (data[ftype, "tangential_%s" % basename] /
                  data[ftype, '%s_magnitude' % basename])
            return np.abs(tr)
        registry.add_field((ftype, "tangential_over_%s_magnitude" % basename),
                           sampling_type="cell",
                           function=_tangential_over_magnitude,
                           take_log=False)

        def _cylindrical_radius_component(field, data):
            """The cylindrical radius component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            theta = data["index", 'cylindrical_theta']
            return get_cyl_r_component(vectors, theta, normal)

        registry.add_field((ftype, "%s_cylindrical_radius" % basename),
                           sampling_type="cell",
                           function=_cylindrical_radius_component,
                           units=field_units,
                           validators=[ValidateParameter("normal")])

        def _cylindrical_radial(field, data):
            """This field is deprecated and will be removed in a future version"""
            return data[ftype, '%s_cylindrical_radius' % basename]

        registry.add_field((ftype, "cylindrical_radial_%s" % basename),
                           sampling_type="cell",
                           function=_cylindrical_radial,
                           units=field_units)

        def _cylindrical_radial_absolute(field, data):
            """This field is deprecated and will be removed in a future version"""
            return np.abs(data[ftype, '%s_cylindrical_radius' % basename])

        registry.add_field((ftype, "cylindrical_radial_%s_absolute" % basename),
                           sampling_type="cell",
                           function=_cylindrical_radial_absolute,
                           units=field_units,
                           validators=[ValidateParameter("normal")])

        def _cylindrical_theta_component(field, data):
            """The cylindrical theta component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            theta = data["index", 'cylindrical_theta'].copy()
            theta = np.tile(theta, (3,) + (1,)*len(theta.shape))
            return get_cyl_theta_component(vectors, theta, normal)

        registry.add_field((ftype, "%s_cylindrical_theta" % basename),
                           sampling_type="cell",
                           function=_cylindrical_theta_component,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center"),
                                       ValidateParameter("bulk_%s" % basename)])

        def _cylindrical_tangential(field, data):
            """This field is deprecated and will be removed in a future release"""
            return data["%s_cylindrical_theta" % basename]

        def _cylindrical_tangential_absolute(field, data):
            """This field is deprecated and will be removed in a future release"""
            return np.abs(data['cylindrical_tangential_%s' % basename])

        registry.add_field((ftype, "cylindrical_tangential_%s" % basename),
                           sampling_type="cell",
                           function=_cylindrical_tangential,
                           units=field_units)

        registry.add_field(
            (ftype, "cylindrical_tangential_%s_absolute" % basename),
            sampling_type="cell",
            function=_cylindrical_tangential_absolute,
            units=field_units)

        def _cylindrical_z_component(field, data):
            """The cylindrical z component of the vector field

            Relative to the coordinate system defined by the *normal* vector,
            *center*, and *bulk_* field parameters.
            """
            normal = data.get_field_parameter("normal")
            vectors = obtain_relative_velocity_vector(
                data, (xn, yn, zn), "bulk_%s" % basename)
            return get_cyl_z_component(vectors, normal)

        registry.add_field((ftype, "%s_cylindrical_z" % basename),
                           sampling_type="cell",
                           function=_cylindrical_z_component,
                           units=field_units,
                           validators=[ValidateParameter("normal"),
                                       ValidateParameter("center"),
                                       ValidateParameter("bulk_%s" % basename)])

    else: # Create Cartesian fields for curvilinear coordinates

        def _cartesian_x(field,data):
          if registry.ds.geometry == "polar":

            return data["%s_r" % basename] * np.cos(data["theta"])

          elif registry.ds.geometry == "cylindrical":

            if data.ds.dimensionality==2:
              return data["%s_r" % basename]
            elif data.ds.dimensionality==3:
              return (data["%s_r" % basename] * np.cos(data["theta"]) -
                      data["%s_theta" % basename] * np.sin(data["theta"]))

          elif registry.ds.geometry == "spherical":

            if data.ds.dimensionality==2:
              return (data["%s_r" % basename] * np.sin(data["theta"]) +
                      data["%s_theta" % basename] * np.cos(data["theta"]))
            elif data.ds.dimensionality==3:
              return (data["%s_r" % basename] * np.sin(data["theta"]) * np.cos(data["phi"]) +
                      data["%s_theta" % basename] * np.cos(data["theta"]) * np.cos(["phi"]) -
                      data["%s_phi" % basename] * np.sin(data["phi"]))

        # it's redundant to define a cartesian x field for 1D data
        if registry.ds.dimensionality > 1:
          registry.add_field((ftype, "%s_cartesian_x" % basename), sampling_type="cell", 
                             function=_cartesian_x,
                             units=field_units,
                             display_field=True)

        def _cartesian_y(field,data):

          if registry.ds.geometry == "polar":

            return data["%s_r" % basename] * np.sin(data["theta"])

          elif registry.ds.geometry == "cylindrical":

            if data.ds.dimensionality==2:
              return data["%s_z" % basename]
            elif data.ds.dimensionality==3:
              return (data["%s_r" % basename] * np.sin(data["theta"]) +
                      data["%s_theta" % basename] * np.cos(data["theta"]))

          elif registry.ds.geometry == "spherical":

            if data.ds.dimensionality==2:
              return (data["%s_r" % basename] * np.cos(data["theta"]) -
                     data["%s_theta" % basename] * np.sin(data["theta"]))
            elif data.ds.dimensionality==3:
              return (data["%s_r" % basename] * np.sin(data["theta"]) * np.sin(data["phi"]) +
                      data["%s_theta" % basename] * np.cos(data["theta"]) * np.sin(["phi"]) +
                      data["%s_phi" % basename] * np.cos(data["phi"]))

        if registry.ds.dimensionality >= 2:
          registry.add_field((ftype, "%s_cartesian_y" % basename), sampling_type="cell", 
                             function=_cartesian_y,
                             units=field_units,
                             display_field=True)

        def _cartesian_z(field,data):

          if registry.ds.geometry == "cylindrical":
            return data["%s_z" % basename]
          elif registry.ds.geometry == "spherical":
            return (data["%s_r" % basename] * np.cos(data["theta"]) -
                    data["%s_theta" % basename] * np.sin(data["theta"]))

        if registry.ds.dimensionality == 3:
          registry.add_field((ftype, "%s_cartesian_z" % basename), sampling_type="cell", 
                             function=_cartesian_z,
                             units=field_units,
                             display_field=True)


def create_averaged_field(registry, basename, field_units, ftype="gas",
                          slice_info=None, validators=None,
                          weight="cell_mass"):

    if validators is None:
        validators = []
    validators += [ValidateSpatial(1, [(ftype, basename)])]

    def _averaged_field(field, data):
        nx, ny, nz = data[(ftype, basename)].shape
        new_field = data.ds.arr(np.zeros((nx-2, ny-2, nz-2), dtype=np.float64),
                                (just_one(data[(ftype, basename)]) *
                                 just_one(data[(ftype, weight)])).units)
        weight_field = data.ds.arr(np.zeros((nx-2, ny-2, nz-2),
                                            dtype=np.float64),
                                   data[(ftype, weight)].units)
        i_i, j_i, k_i = np.mgrid[0:3, 0:3, 0:3]

        for i, j, k in zip(i_i.ravel(), j_i.ravel(), k_i.ravel()):
            sl = (slice(i, nx-(2-i)), slice(j, ny-(2-j)), slice(k, nz-(2-k)))
            new_field += data[(ftype, basename)][sl] * data[(ftype, weight)][sl]
            weight_field += data[(ftype, weight)][sl]

        # Now some fancy footwork
        new_field2 = data.ds.arr(np.zeros((nx, ny, nz)), 
                                 data[(ftype, basename)].units)
        new_field2[1:-1, 1:-1, 1:-1] = new_field / weight_field
        return new_field2

    registry.add_field((ftype, "averaged_%s" % basename),
                       sampling_type="cell",
                       function=_averaged_field,
                       units=field_units,
                       validators=validators)



"""
Definitions specific to ART




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

yt_to_art = {
    'Density': 'HVAR_GAS_DENSITY',
    'TotalEnergy': 'HVAR_GAS_ENERGY',
    'GasEnergy': 'HVAR_INTERNAL_ENERGY',
    'Pressure': 'HVAR_PRESSURE',
    'XMomentumDensity': 'HVAR_MOMENTUM_X',
    'YMomentumDensity': 'HVAR_MOMENTUM_Y',
    'ZMomentumDensity': 'HVAR_MOMENTUM_Z',
    'Gamma': 'HVAR_GAMMA',
    'MetalDensitySNIa': 'HVAR_METAL_DENSITY_Ia',
    'MetalDensitySNII': 'HVAR_METAL_DENSITY_II',
    'Potential': 'VAR_POTENTIAL',
    'PotentialHydro': 'VAR_POTENTIAL_HYDRO',
    'particle_position_x': 'POSITION_X',
    'particle_position_y': 'POSITION_Y',
    'particle_position_z': 'POSITION_Z',
    'particle_velocity_x': 'VELOCITY_X',
    'particle_velocity_y': 'VELOCITY_Y',
    'particle_velocity_z': 'VELOCITY_Z',
    'particle_mass': 'MASS',
    'particle_index': 'PID',
    'particle_species': 'SPECIES',
    'creation_time': 'BIRTH_TIME',
    'particle_mass_initial': 'INITIAL_MASS',
    'particle_metallicity1': 'METALLICITY_SNIa',
    'particle_metallicity2': 'METALLICITY_SNII',
    'stars': 'STAR',
    'nbody': 'N-BODY'
}
art_to_yt = dict(zip(yt_to_art.values(), yt_to_art.keys()))


class ARTIOconstants():
    def __init__(self):
        self.yr = 365.25*86400
        self.Myr = 1.0e6*self.yr
        self.Gyr = 1.0e9*self.yr

        self.pc = 3.0856775813e18
        self.kpc = 1.0e3*self.pc
        self.Mpc = 1.0e6*self.pc

        self.kms = 1.0e5

        self.mp = 1.672621637e-24
        self.k = 1.3806504e-16
        self.G = 6.67428e-8
        self.c = 2.99792458e10

        self.eV = 1.602176487e-12
        self.amu = 1.660538782e-24
        self.mH = 1.007825*self.amu
        self.mHe = 4.002602*self.amu

        self.Msun = 1.32712440018e26/self.G
        self.Zsun = 0.0199

        self.Yp = 0.24
        self.wmu = 4.0/(8.0-5.0*self.Yp)
        self.wmu_e = 1.0/(1.0-0.5*self.Yp)
        self.XH = 1.0 - self.Yp
        self.XHe = 0.25*self.Yp
        self.gamma = 5.0/3.0

        self.sigmaT = 6.6524e-25



import numpy as np

from yt.testing import fake_random_ds, assert_equal
from yt.frontends.stream.data_structures import load_uniform_grid


def test_exclude_above():
    the_ds = fake_random_ds(ndims=3)
    all_data = the_ds.all_data()
    new_ds = all_data.exclude_above('density', 1)
    assert_equal(new_ds['density'], all_data['density'])
    new_ds = all_data.exclude_above('density', 1e6, 'g/m**3')
    assert_equal(new_ds['density'], all_data['density'])
    new_ds = all_data.exclude_above('density', 0)
    assert_equal(new_ds['density'], [])

def test_exclude_below():
    the_ds = fake_random_ds(ndims=3)
    all_data = the_ds.all_data()
    new_ds = all_data.exclude_below('density', 1)
    assert_equal(new_ds['density'], [])
    new_ds = all_data.exclude_below('density', 1e6, 'g/m**3')
    assert_equal(new_ds['density'], [])
    new_ds = all_data.exclude_below('density', 0)
    assert_equal(new_ds['density'], all_data['density'])

def test_exclude_nan():
    test_array = np.nan*np.ones((10, 10, 10))
    test_array[1,1,:] = 1
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape, 
        length_unit='cm', nprocs=1)
    ad = ds.all_data()
    no_nan_ds = ad.exclude_nan('density')
    assert_equal(no_nan_ds['density'], np.array(np.ones(10)))

def test_equal():
    test_array = np.ones((10, 10, 10))
    test_array[1,1,:] = 2.
    test_array[2,1,:] = 3.
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape,
        length_unit='cm', nprocs=1)
    ad = ds.all_data()
    no_ones = ad.exclude_equal('density', 1.0)
    assert np.all(no_ones['density'] != 1.0)
    only_ones = ad.include_equal('density', 1.0)
    assert np.all(only_ones['density'] == 1.0)

def test_inside_outside():
    test_array = np.ones((10, 10, 10))
    test_array[1,1,:] = 2.
    test_array[2,1,:] = 3.
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape,
        length_unit='cm', nprocs=1)
    ad = ds.all_data()
    
    only_ones_and_twos = ad.include_inside('density', 0.9, 2.1)
    assert np.all(only_ones_and_twos['density'] != 3.)
    assert len(only_ones_and_twos['density']) == 990
    
    only_ones_and_twos = ad.exclude_outside('density', 0.9, 2.1)
    assert len(only_ones_and_twos['density']) == 990
    assert np.all(only_ones_and_twos['density'] != 3.)
    
    only_threes = ad.include_outside('density', 0.9, 2.1)
    assert np.all(only_threes['density'] == 3)
    assert len(only_threes['density']) == 10    
    
    only_threes = ad.include_outside('density', 0.9, 2.1)
    assert np.all(only_threes['density'] == 3)
    assert len(only_threes['density']) == 10

    # Repeat, but convert units to g/m**3
    only_ones_and_twos = ad.include_inside('density', 0.9e6, 2.1e6, 'g/m**3')
    assert np.all(only_ones_and_twos['density'] != 3.)
    assert len(only_ones_and_twos['density']) == 990

    only_ones_and_twos = ad.exclude_outside('density', 0.9e6, 2.1e6, 'g/m**3')
    assert len(only_ones_and_twos['density']) == 990
    assert np.all(only_ones_and_twos['density'] != 3.)

    only_threes = ad.include_outside('density', 0.9e6, 2.1e6, 'g/m**3')
    assert np.all(only_threes['density'] == 3)
    assert len(only_threes['density']) == 10

    only_threes = ad.include_outside('density', 0.9e6, 2.1e6, 'g/m**3')
    assert np.all(only_threes['density'] == 3)
    assert len(only_threes['density']) == 10
 



"""
Rockstar-specific fields




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

m_units = "Msun / h"                # Msun / h
p_units = "Mpccm / h"               # Mpc / h comoving
v_units = "km / s"                  # km /s phys, peculiar
r_units = "kpccm / h"               # kpc / h comoving

class RockstarFieldInfo(FieldInfoContainer):
    known_other_fields = (
    )

    known_particle_fields = (
        ("particle_identifier", ("", [], None)),
        ("particle_position_x", (p_units, [], None)),
        ("particle_position_y", (p_units, [], None)),
        ("particle_position_z", (p_units, [], None)),
        ("particle_velocity_x", (v_units, [], None)),
        ("particle_velocity_y", (v_units, [], None)),
        ("particle_velocity_z", (v_units, [], None)),
        ("particle_corevel_x", (v_units, [], None)),
        ("particle_corevel_y", (v_units, [], None)),
        ("particle_corevel_z", (v_units, [], None)),
        ("particle_bulkvel_x", (v_units, [], None)),
        ("particle_bulkvel_y", (v_units, [], None)),
        ("particle_bulkvel_z", (v_units, [], None)),
        ("particle_mass", (m_units, [], "Mass")),
        ("virial_radius", (r_units, [], "Radius")),
        ("child_r", (r_units, [], None)),
        ("vmax_r", (v_units, [], None)),
    # These fields I don't have good definitions for yet.
    ('mgrav', ("", [], None)),
    ('vmax', (v_units, [], "V_{max}")),
    ('rvmax', (v_units, [], None)),
    ('rs', (r_units, [], "R_s")),
    ('klypin_rs', (r_units, [], "Klypin R_s")),
    ('vrms', (v_units, [], "V_{rms}")),
    ('Jx', ("", [], "J_x")),
    ('Jy', ("", [], "J_y")),
    ('Jz', ("", [], "J_z")),
    ('energy', ("", [], None)),
    ('spin', ("", [], "Spin Parameter")),
    ('alt_m1', (m_units, [], None)),
    ('alt_m2', (m_units, [], None)),
    ('alt_m3', (m_units, [], None)),
    ('alt_m4', (m_units, [], None)),
    ('Xoff', ("", [], None)),
    ('Voff', ("", [], None)),
    ('b_to_a', ("", [], "Ellipsoidal b to a")),
    ('c_to_a', ("", [], "Ellipsoidal c to a")),
    ('Ax', ("", [], "A_x")),
    ('Ay', ("", [], "A_y")),
    ('Az', ("", [], "A_z")),
    ('b_to_a2', ("", [], None)),
    ('c_to_a2', ("", [], None)),
    ('A2x', ("", [], "A2_x")),
    ('A2y', ("", [], "A2_y")),
    ('A2z', ("", [], "A2_z")),
    ('bullock_spin', ("", [], "Bullock Spin Parameter")),
    ('kin_to_pot', ("", [], "Kinetic to Potential")),
    ('m_pe_b', ("", [], None)),
    ('m_pe_d', ("", [], None)),
    ('num_p', ("", [], "Number of Particles")),
    ('num_child_particles', ("", [], "Number of Child Particles")),
    ('p_start', ("", [], None)),
    ('desc', ("", [], None)),
    ('flags', ("", [], None)),
    ('n_core', ("", [], None)),
    ('min_pos_err', ("", [], None)),
    ('min_vel_err', ("", [], None)),
    ('min_bulkvel_err', ("", [], None)),
)



"""
Union structures which can be used to form unions of particles, meshes,
etc. Union is the base class from which trivial named union classes
can be derived



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.funcs import ensure_list

class Union(object):
    def __init__(self, name, sub_types):
        self.name = name
        self.sub_types = ensure_list(sub_types)

    def __iter__(self):
        for st in self.sub_types:
            yield st

class MeshUnion(Union):
    def __init__(self, name, sub_types):
        super(MeshUnion, self).__init__(name, sub_types)



import numpy as np
from yt.utilities.exceptions import \
    YTIllDefinedAMR, \
    YTIntDomainOverflow

from yt import load_amr_grids, ProjectionPlot

from yt.testing import assert_raises

def test_qt_overflow():
    grid_data = []

    grid_dict = {}

    grid_dict['left_edge'] = [-1.0, -1.0, -1.0]
    grid_dict['right_edge'] = [1.0, 1.0, 1.0]
    grid_dict['dimensions'] = [8, 8, 8]
    grid_dict['level'] = 0

    grid_dict['density'] = np.ones((8,8,8))

    grid_data.append(grid_dict)

    domain_dimensions = np.array([8, 8, 8])

    spf = load_amr_grids(grid_data, domain_dimensions)

    def make_proj():
        p = ProjectionPlot(spf, 'x', ["density"], center='c', origin='native')
        return p
    assert_raises(YTIntDomainOverflow, make_proj)

def test_refine_by():
    grid_data = []
    ref_by = 4
    lo = 0.0
    hi = 1.0
    fine_grid_width = (hi - lo) / ref_by
    for level in range(2):
        grid_dict = {}

        grid_dict['left_edge'] = [0.0 + 0.5*fine_grid_width*level]*3
        grid_dict['right_edge'] = [1.0 - 0.5*fine_grid_width*level]*3
        grid_dict['dimensions'] = [8, 8, 8]
        grid_dict['level'] = level

        grid_dict['density'] = np.ones((8,8,8))

        grid_data.append(grid_dict)

    domain_dimensions = np.array([8, 8, 8])

    load_amr_grids(grid_data, domain_dimensions, refine_by=ref_by)

def test_validation():
    dims = np.array([4, 2, 4])
    grid_data = [
        dict(left_edge = [0.0, 0.0, 0.0],
             right_edge = [1.0, 1.0, 1.0],
             level = 0,
             dimensions = dims),
        dict(left_edge = [0.25, 0.25, 0.25],
             right_edge = [0.75, 0.75, 0.75],
             level = 1,
             dimensions = dims),
       ]
    bbox = np.array([[0, 1], [0, 1], [0, 1]])
    def load_grids():
        load_amr_grids(grid_data, dims, bbox=bbox, periodicity=(0, 0, 0),
                       length_unit=1.0, refine_by=2)
    assert_raises(YTIllDefinedAMR, load_grids)



"""
Clump tools for use with the yt Clump object



"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
nar = np.array

counter = 0
def recursive_all_clumps(clump,list,level,parentnumber):
    """A recursive function to flatten the index in *clump*.
    Not to be called directly: please call return_all_clumps, below."""
    global counter
    counter += 1
    clump.number = counter
    clump.parentnumber = parentnumber
    counter += 1
    list.append(clump)
    clump.level = level
    if clump.children is not None:
        for child in clump.children:
            recursive_all_clumps(child,list,level+1,clump.number)
    return list

def return_all_clumps(clump):
    """Flatten the index defined by *clump*.
    Additionally adds three variables to the clump:
    level        = depth of index
    number       = index of clump in the final array
    parentnumber = index of this clumps parent

    """
    global counter
    counter = 0
    list = []
    level = 0
    clump.level = level
    parentnumber=-1
    recursive_all_clumps(clump,list,level,parentnumber)
    return list

def return_bottom_clumps(clump,dbg=0):
    """Recursively return clumps at the bottom of the index.
    This gives a list of clumps similar to what one would get from a CLUMPFIND routine"""
    global counter
    counter = 0
    list = []
    level = 0
    recursive_bottom_clumps(clump,list,dbg,level)
    return list
def recursive_bottom_clumps(clump,clump_list, dbg = 0,level=0):
    """Loops over a list of clumps (clumps) and fills clump_list with the bottom most.
    Recursive. Prints the level and the number of cores to the screen."""

    global counter

    if ((clump.children is None) or (len(clump.children) == 0)):
        counter += 1
        clump_list.append( clump )
    else:
        for child in clump.children:
            recursive_bottom_clumps(child,clump_list,dbg=dbg,level=level+1)

def clump_list_sort(clump_list):
    """Returns a copy of clump_list, sorted by ascending minimum density.  This
    eliminates overlap when passing to
    yt.visualization.plot_modification.ClumpContourCallback"""
    minDensity = [c['Density'].min() for c in clump_list]
    
    args = np.argsort(minDensity)
    list = nar(clump_list)[args]
    reverse = range(list.size-1,-1,-1)
    return list[reverse]



# -*- coding: UTF-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import shutil
import sys
import argparse
from yt.config import CURRENT_CONFIG_FILE, _OLD_CONFIG_FILE, YTConfigParser
from yt.extern.six.moves import configparser

CONFIG = YTConfigParser()
CONFIG.read([CURRENT_CONFIG_FILE])


def get_config(section, option):
    return CONFIG.get(section, option)


def set_config(section, option, value):
    if not CONFIG.has_section(section):
        CONFIG.add_section(section)
    CONFIG.set(section, option, value)
    write_config()


def write_config(fd=None):
    if fd is None:
        with open(CURRENT_CONFIG_FILE, 'w') as fd:
            CONFIG.write(fd)
    else:
        CONFIG.write(fd)

def migrate_config():
    if not os.path.exists(_OLD_CONFIG_FILE):
        print('Old config not found.')
        sys.exit()
    CONFIG.read(_OLD_CONFIG_FILE)
    print('Writing a new config file to: {}'.format(CURRENT_CONFIG_FILE))
    write_config()
    print('Backing up the old config file: {}.bak'.format(_OLD_CONFIG_FILE))
    os.rename(_OLD_CONFIG_FILE, _OLD_CONFIG_FILE + '.bak')

    old_config_dir = os.path.dirname(_OLD_CONFIG_FILE)
    try:
        plugin_file = CONFIG.get('yt', 'pluginfilename')
        if plugin_file and \
                os.path.exists(os.path.join(old_config_dir, plugin_file)):
            print('Migrating plugin file {} to new location'.format(plugin_file))
            shutil.copyfile(
                os.path.join(old_config_dir, plugin_file),
                os.path.join(os.path.dirname(CURRENT_CONFIG_FILE), plugin_file))
            print('Backing up the old plugin file: {}.bak'.format(_OLD_CONFIG_FILE))
            plugin_file = os.path.join(old_config_dir, plugin_file)
            os.rename(plugin_file, plugin_file + '.bak')
    except configparser.NoOptionError:
        pass


def rm_config(section, option):
    CONFIG.remove_option(section, option)
    write_config()


def main():
    parser = argparse.ArgumentParser(
        description='Get and set configuration values for yt')
    subparsers = parser.add_subparsers(help='sub-command help', dest='cmd')

    get_parser = subparsers.add_parser('get', help='get a config value')
    set_parser = subparsers.add_parser('set', help='set a config value')
    rm_parser = subparsers.add_parser('rm', help='remove a config option')
    subparsers.add_parser('migrate', help='migrate old config file')
    subparsers.add_parser('list', help='show all config values')

    get_parser.add_argument(
        'section', help='The section containing the option.')
    get_parser.add_argument('option', help='The option to retrieve.')

    set_parser.add_argument(
        'section', help='The section containing the option.')
    set_parser.add_argument('option', help='The option to set.')
    set_parser.add_argument('value', help='The value to set the option to.')

    rm_parser.add_argument(
        'section', help='The section containing the option to remove.')
    rm_parser.add_argument('option', help='The option to remove.')

    args = parser.parse_args()

    if args.cmd == 'get':
        print(get_config(args.section, args.option))
    elif args.cmd == 'set':
        set_config(args.section, args.option, args.value)
    elif args.cmd == 'list':
        write_config(sys.stdout)
    elif args.cmd == 'migrate':
        migrate_config()
    elif args.cmd == 'rm':
        rm_config(args.section, args.option)

if __name__ == '__main__':
    main()  # pragma: no cover
